/*
 * X11 keyboard driver
 *
 * Copyright 1993 Bob Amstadt
 * Copyright 1996 Albrecht Kleine
 * Copyright 1997 David Faure
 * Copyright 1998 Morten Welinder
 * Copyright 1998 Ulrich Weigand
 * Copyright 1999 Ove Kven
 * Copyright 2005 Oleh Nykyforchyn
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 */

#include "config.h"

#include <X11/Xatom.h>
#include <X11/keysym.h>
#include <X11/Xlib.h>
#include <X11/Xresource.h>
#include <X11/Xutil.h>
#ifdef HAVE_XKB
#include <X11/XKBlib.h>
#define MAX_OVERLAYS 4
#define OVER1_ENABLED 1
#define OVER2_ENABLED 2
#define MAX_LAYOUTS (XkbNumKbdGroups*MAX_OVERLAYS)
#define MAX_NEIGHBORS 5
#define MAX_REMEMBER MAX_LAYOUTS
#else
#define MAX_OVERLAYS 1
#define MAX_LAYOUTS 1
#define MAX_NEIGHBORS 5
#define MAX_REMEMBER 10
#endif

#include <ctype.h>
#include <stdarg.h>
#include <string.h>
#include <stdio.h>

#define NONAMELESSUNION
#define NONAMELESSSTRUCT
#include "windef.h"
#include "winbase.h"
#include "wingdi.h"
#include "winuser.h"
#include "wine/winuser16.h"
#include "winnls.h"
#include "winreg.h"
#include "win.h"
#include "x11drv.h"
#include "wine/server.h"
#include "wine/unicode.h"
#include "wine/debug.h"

WINE_DEFAULT_DEBUG_CHANNEL(keyboard); /* Managing the whole layouts set */
WINE_DECLARE_DEBUG_CHANNEL(layout);   /* How each layout is detected */
WINE_DECLARE_DEBUG_CHANNEL(key);      /* Processing single key event */

typedef union
{
    struct
    {
#ifndef BITFIELDS_BIGENDIAN
        unsigned long count : 16;
#endif
        unsigned long code : 8;
        unsigned long extended : 1;
        unsigned long unused : 2;
        unsigned long win_internal : 2;
        unsigned long context : 1;
        unsigned long previous : 1;
        unsigned long transition : 1;
#ifdef BITFIELDS_BIGENDIAN
        unsigned long count : 16;
#endif
    } lp1;
    unsigned long lp2;
} KEYLP;

/* key state table bits:
  0x80 -> key is pressed
  0x40 -> key got pressed since last time
  0x01 -> key is toggled
*/
BYTE key_state_table[256];

static BYTE TrackSysKey = 0; /* determine whether ALT key up will cause a WM_SYSKEYUP
                                or a WM_KEYUP message */

typedef char TempKeycData[2][256][4]; /* To temporarily store
                                         keycode -> char mapping */

typedef struct {
       WORD vkey[256];
       WORD scan[256];
       KeySym ksym[256][4];
} KeycData;

typedef struct {
       int min_keycode;
       int max_keycode;
       int keysyms_per_keycode;
       int num_groups;  /* How many groups are used */
       int num_layouts; /* How many layouts are detected */
       int layout_no; /* current layout, range 0..num_layouts-1 */
       int layout[MAX_LAYOUTS]; /* -1 = refresh  */
       int score[MAX_LAYOUTS];  /* How well the found layout matches */
       int cp[MAX_LAYOUTS];     /* Codepage for translation of keysyms */
       int pref[MAX_LAYOUTS];   /* Scores for preferred layouts (>0), 0 = ignore */
#ifdef HAVE_XKB
       int group; /* Current XKB group */
       int overlay; /* Active overlay groups : 0 = none,
                           1=overlay1, 2=overlay2, 3=both */
       int overlays_used; /* Which overlay groups can be used,
                            values 0,1,2,3 are of the same meaning as above */
       KeyCode keyc2over[MAX_OVERLAYS-1][256]; /* real key mapped to virtual */
       KeyCode over2keyc[MAX_OVERLAYS-1][256]; /* the least real key mapped 
                                                      by overlay to this virtual one */
#endif       
       KeycData* keyc[MAX_LAYOUTS];
       struct {
           int layout; /* Index in main_key_tab, see below; -1=none */
           int score;  /* Numerical measure of how good is the candidate */
           int inv;    /* How many neighbor symbols are swapped */
           int is_raw; /* For explanation see later */
       } prev[MAX_REMEMBER][MAX_NEIGHBORS];
} KbdLayout;

static KbdLayout kbd;

static unsigned global_refresh = 1;     /*   1 = rebuild all tables,
                                             0 = no action */

static int reg_layout[MAX_LAYOUTS]; /* -1 = not specified */
static int reg_cp[MAX_LAYOUTS];     /* -1 = not specified */


#ifdef HAVE_XKB
static XkbDescPtr desc = NULL; /* XKB keyboard description record - here we
                           quickly get all XKB information */
#endif

static const char * INIKeyboard = "Software\\Wine\\X11 Driver\\Keyboard";

static int NumLockMask, AltGrMask; /* mask in the XKeyEvent state */
static int kcControl, kcAlt, kcShift, kcNumLock, kcCapsLock; /* keycodes */

static char KEYBOARD_MapDeadKeysym(KeySym keysym);

/* Keyboard translation tables */
#define MAIN_LEN 49
static const WORD main_key_scan_qwerty[MAIN_LEN] =
{
/* this is my (102-key) keyboard layout, sorry if it doesn't quite match yours */
 /* `    1    2    3    4    5    6    7    8    9    0    -    = */
   0x29,0x02,0x03,0x04,0x05,0x06,0x07,0x08,0x09,0x0A,0x0B,0x0C,0x0D,
 /* q    w    e    r    t    y    u    i    o    p    [    ] */
   0x10,0x11,0x12,0x13,0x14,0x15,0x16,0x17,0x18,0x19,0x1A,0x1B,
 /* a    s    d    f    g    h    j    k    l    ;    '    \ */
   0x1E,0x1F,0x20,0x21,0x22,0x23,0x24,0x25,0x26,0x27,0x28,0x2B,
 /* z    x    c    v    b    n    m    ,    .    / */
   0x2C,0x2D,0x2E,0x2F,0x30,0x31,0x32,0x33,0x34,0x35,
   0x56 /* the 102nd key (actually to the right of l-shift) */
};

static const WORD main_key_scan_abnt_qwerty[MAIN_LEN] =
{
 /* `    1    2    3    4    5    6    7    8    9    0    -    = */
   0x29,0x02,0x03,0x04,0x05,0x06,0x07,0x08,0x09,0x0A,0x0B,0x0C,0x0D,
 /* q    w    e    r    t    y    u    i    o    p    [    ] */
   0x10,0x11,0x12,0x13,0x14,0x15,0x16,0x17,0x18,0x19,0x1A,0x1B,
 /* a    s    d    f    g    h    j    k    l    ;    '    \ */
   0x1E,0x1F,0x20,0x21,0x22,0x23,0x24,0x25,0x26,0x27,0x28,0x2B,
 /* \      z    x    c    v    b    n    m    ,    .    / */
   0x5e,0x2C,0x2D,0x2E,0x2F,0x30,0x31,0x32,0x33,0x34,0x35,
   0x56, /* the 102nd key (actually to the right of l-shift) */
};

static const WORD main_key_scan_dvorak[MAIN_LEN] =
{
 /* `    1    2    3    4    5    6    7    8    9    0    [    ] */
   0x29,0x02,0x03,0x04,0x05,0x06,0x07,0x08,0x09,0x0A,0x0B,0x1A,0x1B,
 /* '    ,    .    p    y    f    g    c    r    l    /    = */
   0x28,0x33,0x34,0x19,0x15,0x21,0x22,0x2E,0x13,0x26,0x35,0x0D,
 /* a    o    e    u    i    d    h    t    n    s    -    \ */
   0x1E,0x18,0x12,0x16,0x17,0x20,0x23,0x14,0x31,0x1F,0x0C,0x2B,
 /* ;    q    j    k    x    b    m    w    v    z */
   0x27,0x10,0x24,0x25,0x2D,0x30,0x32,0x11,0x2F,0x2C,
   0x56 /* the 102nd key (actually to the right of l-shift) */
};

static const WORD main_key_scan_qwerty_jp106[MAIN_LEN] =
{
  /* this is my (106-key) keyboard layout, sorry if it doesn't quite match yours */
 /* 1    2    3    4    5    6    7    8    9    0    -    ^    \ */
   0x02,0x03,0x04,0x05,0x06,0x07,0x08,0x09,0x0A,0x0B,0x0C,0x0D,0x29,
 /* q    w    e    r    t    y    u    i    o    p    @    [ */
   0x10,0x11,0x12,0x13,0x14,0x15,0x16,0x17,0x18,0x19,0x1A,0x1B,
 /* a    s    d    f    g    h    j    k    l    ;    :    ] */
   0x1E,0x1F,0x20,0x21,0x22,0x23,0x24,0x25,0x26,0x27,0x28,0x2B,
 /* z    x    c    v    b    n    m    ,    .    / */
   0x2C,0x2D,0x2E,0x2F,0x30,0x31,0x32,0x33,0x34,0x35,
   0x56 /* the 102nd key (actually to the right of l-shift) */
};


static const WORD main_key_vkey_qwerty[MAIN_LEN] =
{
/* NOTE: this layout must concur with the scan codes layout above */
   VK_OEM_3,'1','2','3','4','5','6','7','8','9','0',VK_OEM_MINUS,VK_OEM_PLUS,
   'Q','W','E','R','T','Y','U','I','O','P',VK_OEM_4,VK_OEM_6,
   'A','S','D','F','G','H','J','K','L',VK_OEM_1,VK_OEM_7,VK_OEM_5,
   'Z','X','C','V','B','N','M',VK_OEM_COMMA,VK_OEM_PERIOD,VK_OEM_2,
   VK_OEM_102 /* the 102nd key (actually to the right of l-shift) */
};

static const WORD main_key_vkey_qwerty_jp106[MAIN_LEN] =
{
/* NOTE: this layout must concur with the scan codes layout above */
   '1','2','3','4','5','6','7','8','9','0',VK_OEM_MINUS,VK_OEM_PLUS,VK_OEM_3,
   'Q','W','E','R','T','Y','U','I','O','P',VK_OEM_4,VK_OEM_6,
   'A','S','D','F','G','H','J','K','L',VK_OEM_1,VK_OEM_7,VK_OEM_5,
   'Z','X','C','V','B','N','M',VK_OEM_COMMA,VK_OEM_PERIOD,VK_OEM_2,
   VK_OEM_102 /* the 102nd key (actually to the right of l-shift) */
};

static const WORD main_key_vkey_qwerty_v2[MAIN_LEN] =
{
/* NOTE: this layout must concur with the scan codes layout above */
   VK_OEM_5,'1','2','3','4','5','6','7','8','9','0',VK_OEM_PLUS,VK_OEM_4,
   'Q','W','E','R','T','Y','U','I','O','P',VK_OEM_6,VK_OEM_1,
   'A','S','D','F','G','H','J','K','L',VK_OEM_3,VK_OEM_7,VK_OEM_2,
   'Z','X','C','V','B','N','M',VK_OEM_COMMA,VK_OEM_PERIOD,VK_OEM_MINUS,
   VK_OEM_102 /* the 102nd key (actually to the right of l-shift) */
};

static const WORD main_key_vkey_qwertz[MAIN_LEN] =
{
/* NOTE: this layout must concur with the scan codes layout above */
   VK_OEM_3,'1','2','3','4','5','6','7','8','9','0',VK_OEM_MINUS,VK_OEM_PLUS,
   'Q','W','E','R','T','Z','U','I','O','P',VK_OEM_4,VK_OEM_6,
   'A','S','D','F','G','H','J','K','L',VK_OEM_1,VK_OEM_7,VK_OEM_5,
   'Y','X','C','V','B','N','M',VK_OEM_COMMA,VK_OEM_PERIOD,VK_OEM_2,
   VK_OEM_102 /* the 102nd key (actually to the right of l-shift) */
};

static const WORD main_key_vkey_qwertz_105[MAIN_LEN] =
{
/* NOTE: this layout must concur with the scan codes layout above */
   VK_OEM_3,'1','2','3','4','5','6','7','8','9','0',VK_OEM_MINUS,VK_OEM_PLUS,
   'Q','W','E','R','T','Z','U','I','O','P',VK_OEM_4,VK_OEM_6,
   'A','S','D','F','G','H','J','K','L',VK_OEM_1,VK_OEM_7,VK_OEM_5,
   VK_OEM_102,'Y','X','C','V','B','N','M',VK_OEM_COMMA,VK_OEM_PERIOD,VK_OEM_2
};

static const WORD main_key_vkey_abnt_qwerty[MAIN_LEN] =
{
/* NOTE: this layout must concur with the scan codes layout above */
   VK_OEM_3,'1','2','3','4','5','6','7','8','9','0',VK_OEM_MINUS,VK_OEM_PLUS,
   'Q','W','E','R','T','Y','U','I','O','P',VK_OEM_4,VK_OEM_6,
   'A','S','D','F','G','H','J','K','L',VK_OEM_1,VK_OEM_8,VK_OEM_5,
   VK_OEM_7,'Z','X','C','V','B','N','M',VK_OEM_COMMA,VK_OEM_PERIOD,VK_OEM_2,
   VK_OEM_102, /* the 102nd key (actually to the right of l-shift) */
};

static const WORD main_key_vkey_azerty[MAIN_LEN] =
{
/* NOTE: this layout must concur with the scan codes layout above */
   VK_OEM_7,'1','2','3','4','5','6','7','8','9','0',VK_OEM_4,VK_OEM_PLUS,
   'A','Z','E','R','T','Y','U','I','O','P',VK_OEM_6,VK_OEM_1,
   'Q','S','D','F','G','H','J','K','L','M',VK_OEM_3,VK_OEM_5,
   'W','X','C','V','B','N',VK_OEM_COMMA,VK_OEM_PERIOD,VK_OEM_2,VK_OEM_8,
   VK_OEM_102 /* the 102nd key (actually to the right of l-shift) */
};

static const WORD main_key_vkey_dvorak[MAIN_LEN] =
{
/* NOTE: this layout must concur with the scan codes layout above */
   VK_OEM_3,'1','2','3','4','5','6','7','8','9','0',VK_OEM_4,VK_OEM_6,
   VK_OEM_7,VK_OEM_COMMA,VK_OEM_PERIOD,'P','Y','F','G','C','R','L',VK_OEM_2,VK_OEM_PLUS,
   'A','O','E','U','I','D','H','T','N','S',VK_OEM_MINUS,VK_OEM_5,
   VK_OEM_1,'Q','J','K','X','B','M','W','V','Z',
   VK_OEM_102 /* the 102nd key (actually to the right of l-shift) */
};

/*** DEFINE YOUR NEW LANGUAGE-SPECIFIC MAPPINGS BELOW, SEE EXISTING TABLES */

/* the VK mappings for the main keyboard will be auto-assigned as before,
   so what we have here is just the character tables */
/* order: Normal, Shift, AltGr, Shift-AltGr */
/* We recommend you write just what is guaranteed to be correct (i.e. what's
   written on the keycaps), not the bunch of special characters behind AltGr
   and Shift-AltGr if it can vary among different X servers */
/* Remember that your 102nd key (to the right of l-shift) should be on a
   separate line, see existing tables */
/* If Wine fails to match your new table, use WINEDEBUG=+layout to find out why */
/* Remember to also add your new table to the layout index table far below! */

/*** German Logitech Desktop Pro keyboard layout */
static const char main_key_DE_logitech[MAIN_LEN][4] =
{
 "^\xb0","1!","2\"","3\xa7","4$","5%","6&","7/{","8([","9)]","0=}","\xdf?\\","'`",
 "qQ@","wW","eE","rR","tT","zZ","uU","iI","oO","pP","\xfc\xdc","+*~",
 "aA","sS","dD","fF","gG","hH","jJ","kK","lL","\xf6\xd6","\xe4\xc4","#'",
 "yY","xX","cC","vV","bB","nN","mM",",;",".:","-_",
 "<>|"
};

/*** United States keyboard layout (mostly contributed by Uwe Bonnes) */
static const char main_key_US[MAIN_LEN][4] =
{
 "`~","1!","2@","3#","4$","5%","6^","7&","8*","9(","0)","-_","=+",
 "qQ","wW","eE","rR","tT","yY","uU","iI","oO","pP","[{","]}",
 "aA","sS","dD","fF","gG","hH","jJ","kK","lL",";:","'\"","\\|",
 "zZ","xX","cC","vV","bB","nN","mM",",<",".>","/?"
};

/*** United States keyboard layout (phantom key version) */
/* (XFree86 reports the <> key even if it's not physically there) */
static const char main_key_US_phantom[MAIN_LEN][4] =
{
 "`~","1!","2@","3#","4$","5%","6^","7&","8*","9(","0)","-_","=+",
 "qQ","wW","eE","rR","tT","yY","uU","iI","oO","pP","[{","]}",
 "aA","sS","dD","fF","gG","hH","jJ","kK","lL",";:","'\"","\\|",
 "zZ","xX","cC","vV","bB","nN","mM",",<",".>","/?",
 "<>" /* the phantom key */
};

/*** United States keyboard layout (dvorak version) */
static const char main_key_US_dvorak[MAIN_LEN][4] =
{
 "`~","1!","2@","3#","4$","5%","6^","7&","8*","9(","0)","[{","]}",
 "'\"",",<",".>","pP","yY","fF","gG","cC","rR","lL","/?","=+",
 "aA","oO","eE","uU","iI","dD","hH","tT","nN","sS","-_","\\|",
 ";:","qQ","jJ","kK","xX","bB","mM","wW","vV","zZ"
};

/*** British keyboard layout */
static const char main_key_UK[MAIN_LEN][4] =
{
 "`","1!","2\"","3","4$","5%","6^","7&","8*","9(","0)","-_","=+",
 "qQ","wW","eE","rR","tT","yY","uU","iI","oO","pP","[{","]}",
 "aA","sS","dD","fF","gG","hH","jJ","kK","lL",";:","'@","#~",
 "zZ","xX","cC","vV","bB","nN","mM",",<",".>","/?",
 "\\|"
};

/*** French keyboard layout (setxkbmap fr) */
static const char main_key_FR[MAIN_LEN][4] =
{
 "","&1","2","\"3","'4","(5","-6","7","_8","9","0",")","=+",
 "aA","zZ","eE","rR","tT","yY","uU","iI","oO","pP","^","$",
 "qQ","sS","dD","fF","gG","hH","jJ","kK","lL","mM","%","*",
 "wW","xX","cC","vV","bB","nN",",?",";.",":/","!",
 "<>"
};

/*** Icelandic keyboard layout (setxkbmap is) */
static const char main_key_IS[MAIN_LEN][4] =
{
 "","1!","2\"","3#","4$","5%","6&","7/","8(","9)","0=","","-_",
 "qQ","wW","eE","rR","tT","yY","uU","iI","oO","pP","","'?",
 "aA","sS","dD","fF","gG","hH","jJ","kK","lL","","","+*",
 "zZ","xX","cC","vV","bB","nN","mM",",;",".:","",
 "<>"
};

/*** German keyboard layout (setxkbmap de) */
static const char main_key_DE[MAIN_LEN][4] =
{
 "^","1!","2\"","3","4$","5%","6&","7/","8(","9)","0=","?","`",
 "qQ","wW","eE","rR","tT","zZ","uU","iI","oO","pP","","+*",
 "aA","sS","dD","fF","gG","hH","jJ","kK","lL","","","#'",
 "yY","xX","cC","vV","bB","nN","mM",",;",".:","-_",
 "<>|"
};

/*** German keyboard layout without dead keys */
static const char main_key_DE_nodead[MAIN_LEN][4] =
{
 "^","1!","2\"","3","4$","5%","6&","7/{","8([","9)]","0=}","?\\","",
 "qQ","wW","eE","rR","tT","zZ","uU","iI","oO","pP","","+*~",
 "aA","sS","dD","fF","gG","hH","jJ","kK","lL","","","#'",
 "yY","xX","cC","vV","bB","nN","mM",",;",".:","-_",
 "<>"
};

/*** German keyboard layout without dead keys 105 Keys (contributed by Matthias Fechner)*/
static const char main_key_DE_nodead_105[MAIN_LEN][4] =
{
 "^","1!","2\"","3","4$","5%","6&","7/{","8([","9)]","0=}","?\\","'`",
 "qQ@","wW","eE","rR","tT","zZ","uU","iI","oO","pP","","+*~",
 "aA","sS","dD","fF","gG","hH","jJ","kK","lL","","","#'",
 "<>|","yY","xX","cC","vV","bB","nN","mM",",;",".:","-_",
};

/*** German keyboard layout (simple, minimal difference from US layout) */
static const char main_key_DE_std[MAIN_LEN][4] =
{
 "\xdf\xb0","1!","2@","3#","4$","5%","6^","7&","8*","9(","0)","-_","=+",
 "qQ","wW","eE","rR","tT","yY","uU","iI","oO","pP","","`'~",
 "aA","sS","dD","fF","gG","hH","jJ","kK","lL","","","\\|",
 "zZ","xX","cC","vV","bB","nN","mM",",<",".>","/?",
 "<>" /* the phantom key */
};

/*** Swiss German keyboard layout (setxkbmap ch -variant de) */
static const char main_key_SG[MAIN_LEN][4] =
{
 "","1+","2\"","3*","4","5%","6&","7/","8(","9)","0=","'?","^`",
 "qQ","wW","eE","rR","tT","zZ","uU","iI","oO","pP","","!",
 "aA","sS","dD","fF","gG","hH","jJ","kK","lL","","","$",
 "yY","xX","cC","vV","bB","nN","mM",",;",".:","-_",
 "<>"
};

/*** Swiss French keyboard layout (setxkbmap ch -variant fr) */
static const char main_key_SF[MAIN_LEN][4] =
{
 "","1+","2\"","3*","4","5%","6&","7/","8(","9)","0=","'?","^`",
 "qQ","wW","eE","rR","tT","zZ","uU","iI","oO","pP","","!",
 "aA","sS","dD","fF","gG","hH","jJ","kK","lL","","","$",
 "yY","xX","cC","vV","bB","nN","mM",",;",".:","-_",
 "<>"
};

/*** Norwegian keyboard layout (contributed by Ove Kven) */
static const char main_key_NO[MAIN_LEN][4] =
{
 "|","1!","2\"@","3#","4$","5%","6&","7/{","8([","9)]","0=}","+?","\\`",
 "qQ","wW","eE","rR","tT","yY","uU","iI","oO","pP","","^~",
 "aA","sS","dD","fF","gG","hH","jJ","kK","lL","","","'*",
 "zZ","xX","cC","vV","bB","nN","mM",",;",".:","-_",
 "<>"
};

/*** Danish keyboard layout (setxkbmap dk) */
static const char main_key_DA[MAIN_LEN][4] =
{
 "","1!","2\"","3#","4","5%","6&","7/","8(","9)","0=","+?","`",
 "qQ","wW","eE","rR","tT","yY","uU","iI","oO","pP","","^",
 "aA","sS","dD","fF","gG","hH","jJ","kK","lL","","","'*",
 "zZ","xX","cC","vV","bB","nN","mM",",;",".:","-_",
 "<>"
};

/*** Swedish keyboard layout (setxkbmap se) */
static const char main_key_SE[MAIN_LEN][4] =
{
 "","1!","2\"","3#","4","5%","6&","7/","8(","9)","0=","+?","`",
 "qQ","wW","eE","rR","tT","yY","uU","iI","oO","pP","","^",
 "aA","sS","dD","fF","gG","hH","jJ","kK","lL","","","'*",
 "zZ","xX","cC","vV","bB","nN","mM",",;",".:","-_",
 "<>"
};

/*** Estonian keyboard layout (setxkbmap ee) */
static const char main_key_ET[MAIN_LEN][4] =
{
 "~","1!","2\"","3#","4","5%","6&","7/","8(","9)","0=","+?","`",
 "qQ","wW","eE","rR","tT","yY","uU","iI","oO","pP","","",
 "aA","sS","dD","fF","gG","hH","jJ","kK","lL","","","'*",
 "zZ","xX","cC","vV","bB","nN","mM",",;",".:","-_",
 "<>"
};

/*** Canadian French keyboard layout (setxkbmap ca_enhanced) */
static const char main_key_CF[MAIN_LEN][4] =
{
 "#|\\","1!","2\"@","3/","4$","5%","6?","7&","8*","9(","0)","-_","=+",
 "qQ","wW","eE","rR","tT","yY","uU","iI","oO","pP","^^[","]",
 "aA","sS","dD","fF","gG","hH","jJ","kK","lL",";:~","``{","<>}",
 "zZ","xX","cC","vV","bB","nN","mM",",'-",".","",
 ""
};

/*** Canadian French keyboard layout (setxkbmap ca -variant fr) */
static const char main_key_CA_fr[MAIN_LEN][4] =
{
 "#|","1!","2\"","3/","4$","5%","6?","7&","8*","9(","0)","-_","=+",
 "qQ","wW","eE","rR","tT","yY","uU","iI","oO","pP","^^","",
 "aA","sS","dD","fF","gG","hH","jJ","kK","lL",";:","``","<>",
 "zZ","xX","cC","vV","bB","nN","mM",",'",".","",
 ""
};

/*** Canadian keyboard layout (setxkbmap ca) */
static const char main_key_CA[MAIN_LEN][4] =
{
 "/\\","1!","2@","3#","4$","5%","6?","7&","8*","9(","0)","-_","=+",
 "qQ","wW","eE","rR","tT","yY","uU","iI","oO","pP","^","~",
 "aA","sSߧ","dD","fF","gG","hH","jJ","kK","lL",";:","","",
 "zZ","xX","cC","vV","bB","nN","mM",",'",".\"","",
 ""
};

/*** Portuguese keyboard layout (setxkbmap pt) */
static const char main_key_PT[MAIN_LEN][4] =
{
 "\\|","1!","2\"","3#","4$","5%","6&","7/","8(","9)","0=","'?","",
 "qQ","wW","eE","rR","tT","yY","uU","iI","oO","pP","+*","`",
 "aA","sS","dD","fF","gG","hH","jJ","kK","lL","","","~^",
 "zZ","xX","cC","vV","bB","nN","mM",",;",".:","-_",
 "<>"
};

/*** Italian keyboard layout (setxkbmap it) */
static const char main_key_IT[MAIN_LEN][4] =
{
 "\\|","1!","2\"","3","4$","5%","6&","7/","8(","9)","0=","'?","^",
 "qQ","wW","eE","rR","tT","yY","uU","iI","oO","pP","","+*",
 "aA","sS","dD","fF","gG","hH","jJ","kK","lL","","","",
 "zZ","xX","cC","vV","bB","nN","mM",",;",".:","-_",
 "<>"
};

/*** Finnish keyboard layout (setxkbmap fi) */
static const char main_key_FI[MAIN_LEN][4] =
{
 "","1!","2\"","3#","4","5%","6&","7/","8(","9)","0=","+?","`",
 "qQ","wW","eE","rR","tT","yY","uU","iI","oO","pP","","^",
 "aA","sS","dD","fF","gG","hH","jJ","kK","lL","","","'*",
 "zZ","xX","cC","vV","bB","nN","mM",",;",".:","-_",
 "<>"
};

/*** Bulgarian bds keyboard layout */
static const char main_key_BG_bds[MAIN_LEN][4] =
{
 "`~()","1!","2@2?","3#3+","4$4\"","5%","6^6=","7&7:","8*8/","9(","0)","-_-I","=+.V",
 "qQ,","wW","eE","rR","tT","yY","uU","iI","oO","pP","[{","]};",
 "aA","sS","dD","fF","gG","hH","jJ","kK","lL",";:","'\"","\\|'",
 "zZ","xX","cC","vV","bB","nN","mM",",<",".>","/?",
 "<>" /* the phantom key */
};

/*** Bulgarian phonetic keyboard layout */
static const char main_key_BG_phonetic[MAIN_LEN][4] =
{
 "`~","1!","2@","3#","4$","5%","6^","7&","8*","9(","0)","-_","=+",
 "qQ","wW","eE","rR","tT","yY","uU","iI","oO","pP","[{","]}",
 "aA","sS","dD","fF","gG","hH","jJ","kK","lL",";:","'\"","\\|",
 "zZ","xX","cC","vV","bB","nN","mM",",<",".>","/?",
 "<>" /* the phantom key */
};

/*** Belarusian standard keyboard layout (contributed by Hleb Valoska) */
/*** It matches belarusian layout for XKB from Alexander Mikhailian    */
static const char main_key_BY[MAIN_LEN][4] =
{
 "`~","1!","2@","3#","4$","5%","6^","7&","8*","9(","0)","-_","=+",
 "qQ","wW","eE","rR","tT","yY","uU","iI","oO","pP","[{","]}''",
 "aA","sS","dD","fF","gG","hH","jJ","kK","lL",";:","'\"","\\|/|",
 "zZ","xX","cC","vV","bB","nN","mM",",<",".>","/?.,", "<>|",
};


/*** Russian keyboard layout (contributed by Pavel Roskin) */
static const char main_key_RU[MAIN_LEN][4] =
{
 "`~","1!","2@","3#","4$","5%","6^","7&","8*","9(","0)","-_","=+",
 "qQ","wW","eE","rR","tT","yY","uU","iI","oO","pP","[{","]}",
 "aA","sS","dD","fF","gG","hH","jJ","kK","lL",";:","'\"","\\|",
 "zZ","xX","cC","vV","bB","nN","mM",",<",".>","/?"
};

/*** Russian keyboard layout (phantom key version) */
static const char main_key_RU_phantom[MAIN_LEN][4] =
{
 "`~","1!","2@","3#","4$","5%","6^","7&","8*","9(","0)","-_","=+",
 "qQ","wW","eE","rR","tT","yY","uU","iI","oO","pP","[{","]}",
 "aA","sS","dD","fF","gG","hH","jJ","kK","lL",";:","'\"","\\|",
 "zZ","xX","cC","vV","bB","nN","mM",",<",".>","/?",
 "<>" /* the phantom key */
};

/*** Russian keyboard layout KOI8-R */
static const char main_key_RU_koi8r[MAIN_LEN][4] =
{
 "()","1!","2\"","3/","4$","5:","6,","7.","8;","9?","0%","-_","=+",
 "","","","","","","","","","","","",
 "","","","","","","","","","","","\\|",
 "","","","","","","","","","/?",
 "<>" /* the phantom key */
};

/*** Russian keyboard layout cp1251 */
static const char main_key_RU_cp1251[MAIN_LEN][4] =
{
 "`~","1!","2@","3#","4$","5%","6^","7&","8*","9(","0)","-_","=+",
 "qQ","wW","eE","rR","tT","yY","uU","iI","oO","pP","[{","]}",
 "aA","sS","dD","fF","gG","hH","jJ","kK","lL",";:","'\"","\\|",
 "zZ","xX","cC","vV","bB","nN","mM",",<",".>","/?",
 "<>" /* the phantom key */
};

/*** Russian phonetic keyboard layout */
static const char main_key_RU_phonetic[MAIN_LEN][4] =
{
 "`~","1!","2@","3#","4$","5%","6^","7&","8*","9(","0)","-_","=+",
 "qQ","wW","eE","rR","tT","yY","uU","iI","oO","pP","[{","]}",
 "aA","sS","dD","fF","gG","hH","jJ","kK","lL",";:","'\"","\\|",
 "zZ","xX","cC","vV","bB","nN","mM",",<",".>","/?",
 "<>" /* the phantom key */
};

/*** Ukrainian keyboard layout KOI8-U */
static const char main_key_UA[MAIN_LEN][4] =
{
 "`~","1!1!","2@2\"","3#3'","4$4*","5%5:","6^6,","7&7.","8*8;","9(9(","0)0)","-_-_","=+=+",
 "qQ","wW","eE","rR","tT","yY","uU","iI","oO","pP","[{","]}",
 "aA","sS","dD","fF","gG","hH","jJ","kK","lL",";:","'\"","\\|\\|",
 "zZ","xX","cC","vV","bB","nN","mM",",<",".>","/?/?",
 "<>" /* the phantom key */
};

/*** Ukrainian keyboard layout KOI8-U by O. Nykyforchyn */
/***  (as it appears on most of keyboards sold today)   */
static const char main_key_UA_std[MAIN_LEN][4] =
{
 "","1!","2\"","3'","4;","5%","6:","7?","8*","9(","0)","-_","=+",
 "","","","","","","","","","","","",
 "","","","","","","","","","","","\\/",
 "","","","","","","","","",".,",
 "<>" /* the phantom key */
};

/*** Russian keyboard layout KOI8-R (pair to the previous) */
static const char main_key_RU_std[MAIN_LEN][4] =
{
 "","1!","2\"","3'","4;","5%","6:","7?","8*","9(","0)","-_","=+",
 "","","","","","","","","","","","",
 "","","","","","","","","","","","\\/",
 "","","","","","","","","",".,",
 "<>" /* the phantom key */
};

/*** Belarusian keyboard layout KOI8-RU */
static const char main_key_BY_std[MAIN_LEN][4] =
{
 "","1!","2\"","3'","4;","5%","6:","7?","8*","9(","0)","-_","=+",
 "","","","","","","","","","","","",
 "","","","","","","","","","","","\\/",
 "","","","","","","","","",".,",
 "<>" /* the phantom key */
};

/*** Spanish keyboard layout (setxkbmap es) */
static const char main_key_ES[MAIN_LEN][4] =
{
 "","1!","2\"","3","4$","5%","6&","7/","8(","9)","0=","'?","",
 "qQ","wW","eE","rR","tT","yY","uU","iI","oO","pP","`^","+*",
 "aA","sS","dD","fF","gG","hH","jJ","kK","lL","","","",
 "zZ","xX","cC","vV","bB","nN","mM",",;",".:","-_",
 "<>"
};

/*** Belgian keyboard layout ***/
static const char main_key_BE[MAIN_LEN][4] =
{
 "","&1|","2@","\"3#","'4","(5","6^","7","!8","9{","0}",")","-_",
 "aA","zZ","eE","rR","tT","yY","uU","iI","oO","pP","^[","$*]",
 "qQ","sS","dD","fF","gG","hH","jJ","kK","lL","mM","%","`",
 "wW","xX","cC","vV","bB","nN",",?",";.",":/","=+~",
 "<>\\"
};

/*** Hungarian keyboard layout (setxkbmap hu) */
static const char main_key_HU[MAIN_LEN][4] =
{
 "0","1'","2\"","3+","4!","5%","6/","7=","8(","9)","","","",
 "qQ","wW","eE","rR","tT","zZ","uU","iI","oO","pP","","",
 "aA","sS","dD","fF","gG","hH","jJ","kK","lL","","","",
 "yY","xX","cC","vV","bB","nN","mM",",?",".:","-_",
 ""
};

/*** Polish (programmer's) keyboard layout ***/
static const char main_key_PL[MAIN_LEN][4] =
{
 "`~","1!","2@","3#","4$","5%","6^","7&","8*","9(","0)","-_","=+",
 "qQ","wW","eE","rR","tT","yY","uU","iI","oO","pP","[{","]}",
 "aA","sS","dD","fF","gG","hH","jJ","kK","lL",";:","'\"","\\|",
 "zZ","xX","cC","vV","bB","nN","mM",",<",".>","/?",
 "<>|"
};

/*** Polish keyboard layout (simple, minimal difference from US layout) */
static const char main_key_PL_std[MAIN_LEN][4] =
{
 "`~","1!","2@","3#","4$","5%","6^","7&","8*","9(","0)","-_","=+",
 "qQ","wW","eE","rR","tT","yY","uU","iI","oO","pP","\xea\xca","\xb3\xa3",
 "aA","sS","dD","fF","gG","hH","jJ","kK","lL","\xf3\xd3","\xb1\xa1","\\|",
 "zZ","xX","cC","vV","bB","nN","mM",",<",".>","/?",
 "<>" /* the phantom key */
};

/*** Slovenian keyboard layout (setxkbmap si) ***/
static const char main_key_SI[MAIN_LEN][4] =
{
 "","1!","2\"","3#","4$","5%","6&","7/","8(","9)","0=","'?","+*",
 "qQ","wW","eE","rR","tT","zZ","uU","iI","oO","pP","","",
 "aA","sS","dD","fF","gG","hH","jJ","kK","lL","","","",
 "yY","xX","cC","vV","bB","nN","mM",",;",".:","-_",
 "<>"
};

/*** Serbian keyboard layout (setxkbmap sr) ***/
static const char main_key_SR[MAIN_LEN][4] =
{
 "`~","1!","2\"","3#","4$","5%","6&","7/","8(","9)","0=","'?","+*",
 "","","","","","","","","","","","[]",
 "","","","","","","","","","","","-_",
 "","","","","","","",",;",".:","",
 "<>" /* the phantom key */
};

/*** Serbian keyboard layout (setxkbmap us,sr) ***/
static const char main_key_US_SR[MAIN_LEN][4] =
{
 "`~","1!","2@2\"","3#","4$","5%","6^6&","7&7/","8*8(","9(9)","0)0=","-_'?","=++*",
 "qQ","wW","eE","rR","tT","yY","uU","iI","oO","pP","[{","]}[]",
 "aA","sS","dD","fF","gG","hH","jJ","kK","lL",";:","'\"","\\|-_",
 "zZ","xX","cC","vV","bB","nN","mM",",<,;",".>.:","/?",
 "<>" /* the phantom key */
};

/*** Croatian keyboard layout specific for me <jelly@srk.fer.hr> ***/
static const char main_key_HR_jelly[MAIN_LEN][4] =
{
 "`~","1!","2@","3#","4$","5%","6^","7&","8*","9(","0)","-_","=+",
 "qQ","wW","eE","rR","tT","yY","uU","iI","oO","pP","[{","]}",
 "aA","sS","dD","fF","gG","hH","jJ","kK","lL",";:","'\"","\\|",
 "zZ","xX","cC","vV","bB","nN","mM",",<",".>","/?",
 "<>|"
};

/*** Croatian keyboard layout (setxkbmap hr) ***/
static const char main_key_HR[MAIN_LEN][4] =
{
 "","1!","2\"","3#","4$","5%","6&","7/","8(","9)","0=","'?","+*",
 "qQ","wW","eE","rR","tT","zZ","uU","iI","oO","pP","","",
 "aA","sS","dD","fF","gG","hH","jJ","kK","lL","","","",
 "yY","xX","cC","vV","bB","nN","mM",",;",".:","/?",
 "<>"
};

/*** Japanese 106 keyboard layout ***/
static const char main_key_JA_jp106[MAIN_LEN][4] =
{
 "1!","2\"","3#","4$","5%","6&","7'","8(","9)","0~","-=","^~","\\|",
 "qQ","wW","eE","rR","tT","yY","uU","iI","oO","pP","@`","[{",
 "aA","sS","dD","fF","gG","hH","jJ","kK","lL",";+",":*","]}",
 "zZ","xX","cC","vV","bB","nN","mM",",<",".>","/?",
 "\\_",
};

/*** Japanese pc98x1 keyboard layout ***/
static const char main_key_JA_pc98x1[MAIN_LEN][4] =
{
 "1!","2\"","3#","4$","5%","6&","7'","8(","9)","0","-=","^`","\\|",
 "qQ","wW","eE","rR","tT","yY","uU","iI","oO","pP","@~","[{",
 "aA","sS","dD","fF","gG","hH","jJ","kK","lL",";+",":*","]}",
 "zZ","xX","cC","vV","bB","nN","mM",",<",".>","/?",
 "\\_",
};

/*** Brazilian ABNT-2 keyboard layout (contributed by Raul Gomes Fernandes) */
static const char main_key_PT_br[MAIN_LEN][4] =
{
 "'\"","1!","2@","3#","4$","5%","6","7&","8*","9(","0)","-_","=+",
 "qQ","wW","eE","rR","tT","yY","uU","iI","oO","pP","`","[{",
 "aA","sS","dD","fF","gG","hH","jJ","kK","lL","","~^","]}",
 "\\|","zZ","xX","cC","vV","bB","nN","mM",",<",".>",";:","/?",
};

/*** Brazilian ABNT-2 keyboard layout with <ALT GR> (contributed by Mauro Carvalho Chehab) */
static const char main_key_PT_br_alt_gr[MAIN_LEN][4] =
{
 "'\"","1!9","2@2","3#3","4$#","5%\"","6(,","7&","8*","9(","0)","-_","=+'",
 "qQ","wW","eE","rR","tT","yY","uU","iI","oO","pP","4`","[{*",
 "aA","sS","dD","fF","gG","hH","jJ","kK","lL","gG","~^","]}:",
 "\\|","zZ","xX","cC","vV","bB","nN","mM",",<",".>",";:","/?0"
};

/*** US international keyboard layout (contributed by Gustavo Noronha (kov@debian.org)) */
static const char main_key_US_intl[MAIN_LEN][4] =
{
  "`~", "1!", "2@", "3#", "4$", "5%", "6^", "7&", "8*", "9(", "0)", "-_", "=+", "\\|",
  "qQ", "wW", "eE", "rR", "tT", "yY", "uU", "iI", "oO", "pP", "[{", "]}",
  "aA", "sS", "dD", "fF", "gG", "hH", "jJ", "kK", "lL", ";:", "'\"",
  "zZ", "xX", "cC", "vV", "bB", "nN", "mM", ",<", ".>", "/?"
};

/*** Slovak keyboard layout (see cssk_ibm(sk_qwerty) in xkbsel)
  - dead_abovering replaced with degree - no symbol in iso8859-2
  - brokenbar replaced with bar					*/
static const char main_key_SK[MAIN_LEN][4] =
{
 ";0","+1","2","3","4","5","6","7","8","9","0","=%","'v",
 "qQ","wW","eE","rR","tT","yY","uU","iI","oO","pP","/","(",
 "aA","sS","dD","fF","gG","hH","jJ","kK","lL","\"","!",")",
 "zZ","xX","cC","vV","bB","nN","mM",",?",".:","-_",
 "<>"
};

/*** Czech keyboard layout (setxkbmap cz) */
static const char main_key_CZ[MAIN_LEN][4] =
{
 ";","+1","2","3","4","5","6","7","8","9","0","=%","",
 "qQ","wW","eE","rR","tT","zZ","uU","iI","oO","pP","/",")(",
 "aA","sS","dD","fF","gG","hH","jJ","kK","lL","\"","!","'",
 "yY","xX","cC","vV","bB","nN","mM",",?",".:","-_",
 "\\"
};

/*** Czech keyboard layout (setxkbmap cz_qwerty) */
static const char main_key_CZ_qwerty[MAIN_LEN][4] =
{
 ";","+1","2","3","4","5","6","7","8","9","0","=%","",
 "qQ","wW","eE","rR","tT","yY","uU","iI","oO","pP","/",")(",
 "aA","sS","dD","fF","gG","hH","jJ","kK","lL","\"","!","'",
 "zZ","xX","cC","vV","bB","nN","mM",",?",".:","-_",
 "\\"
};

/*** Slovak and Czech (programmer's) keyboard layout (see cssk_dual(cs_sk_ucw)) */
static const char main_key_SK_prog[MAIN_LEN][4] =
{
 "`~","1!","2@","3#","4$","5%","6^","7&","8*","9(","0)","-_","=+",
 "qQ","wW","eE","rR","tT","yY","uU","iI","oO","pP","[{","]}",
 "aA","sS","dD","fF","gG","hH","jJ","kK","lL",";:","'\"","\\|",
 "zZ","xX","cC","vV","bB","nN","mM",",<",".>","/?",
 "<>"
};

/*** Czech keyboard layout (see cssk_ibm(cs_qwerty) in xkbsel) */
static const char main_key_CS[MAIN_LEN][4] =
{
 ";","+1","2","3","4","5","6","7","8","9","0)","=%","",
 "qQ\\","wW|","eE","rR","tT","yY","uU","iI","oO","pP","/[{",")(]}",
 "aA","sS","dD","fF[","gG]","hH","jJ","kK","lL","\"$","!","'",
 "zZ>","xX#","cC&","vV@","bB{","nN}","mM",",?<",".:>","-_*",
 "<>\\|"
};

/*** Latin American keyboard layout (contributed by Gabriel Orlando Garcia) */
static const char main_key_LA[MAIN_LEN][4] =
{
 "|","1!","2\"","3#","4$","5%","6&","7/","8(","9)","0=","'?","",
 "qQ@","wW","eE","rR","tT","yY","uU","iI","oO","pP","","+*",
 "aA","sS","dD","fF","gG","hH","jJ","kK","lL","","{[^","}]",
 "zZ","xX","cC","vV","bB","nN","mM",",;",".:","-_",
 "<>"
};

/*** Lithuanian keyboard layout (setxkbmap lt) */
static const char main_key_LT_B[MAIN_LEN][4] =
{
 "`~","","","","","","","","","(",")","-_","","\\|",
 "qQ","wW","eE","rR","tT","yY","uU","iI","oO","pP","[{","]}",
 "aA","sS","dD","fF","gG","hH","jJ","kK","lL",";:","'\"",
 "zZ","xX","cC","vV","bB","nN","mM",",<",".>","/?",
 ""
};

/*** Turkish keyboard Layout */
static const char main_key_TK[MAIN_LEN][4] =
{
"\"","1!","2'","3^#","4+$","5%","6&","7/{","8([","9)]","0=}","*?\\","-_",
"qQ@","wW","eE","rR","tT","yY","uU","I","oO","pP","","~",
"aA","sS","dD","fF","gG","hH","jJ","kK","lL","","i",",;`",
"zZ","xX","cC","vV","bB","nN","mM","","",".:"
};

/*** Turkish keyboard layout (setxkbmap tr) */
static const char main_key_TR[MAIN_LEN][4] =
{
"\"\\","1!","2'","3^","4+","5%","6&","7/","8(","9)","0=","*?","-_",
"qQ","wW","eE","rR","tT","yY","uU","\xb9I","oO","pP","\xbb\xab","",
"aA","sS","dD","fF","gG","hH","jJ","kK","lL","\xba\xaa","i\0",",;",
"zZ","xX","cC","vV","bB","nN","mM","","",".:",
"<>"
};

/*** Turkish F keyboard layout (setxkbmap trf) */
static const char main_key_TR_F[MAIN_LEN][4] =
{
"+*","1!","2\"","3^#","4$","5%","6&","7'","8(","9)","0=","/?","-_",
"fF","gG","\xbb\xab","\xb9I","oO","dD","rR","nN","hH","pP","qQ","wW",
"uU","i\0","eE","aA","","tT","kK","mM","lL","yY","\xba\xaa","xX",
"jJ","","vV","cC","","zZ","sS","bB",".:",",;",
"<>"
};

/*** Israelian keyboard layout (setxkbmap us,il) */
static const char main_key_IL[MAIN_LEN][4] =
{
 "`~;","1!","2@","3#","4$","5%","6^","7&","8*","9(","0)","-_","=+",
 "qQ/","wW'","eE","rR","tT","yY","uU","iI","oO","pP","[{","]}",
 "aA","sS","dD","fF","gG","hH","jJ","kK","lL",";:","\'\",","\\|",
 "zZ","xX","cC","vV","bB","nN","mM",",<",".>","/?.",
 "<>"
};

/*** Israelian phonetic keyboard layout (setxkbmap us,il_phonetic) */
static const char main_key_IL_phonetic[MAIN_LEN][4] =
{
 "`~","1!","2@","3#","4$","5%","6^","7&","8*","9(","0)","-_","=+",
 "qQ","wW","eE","rR","tT","yY","uU","iI","oO","pP","[{","]}",
 "aA","sS","dD","fF","gG","hH","jJ","kK","lL",";:","'\"","\\|",
 "zZ","xX","cC","vV","bB","nN","mM",",<",".>","/?",
 "<>"
};

/*** Israelian Saharon keyboard layout (setxkbmap -symbols "us(pc105)+il_saharon") */
static const char main_key_IL_saharon[MAIN_LEN][4] =
{
 "`~","1!","2@","3#","4$","5%","6^","7&","8*","9(","0)","-_","=+",
 "qQ","wW","eE","rR","tT","yY","uU","iI","oO","pP","[{","]}",
 "aA","sS","dD","fF","gG","hH","jJ","kK","lL",";:","'\"","\\|",
 "zZ","xX","cC","vV","bB","nN","mM",",<",".>","/?",
 "<>"
};

/*** Greek keyboard layout (contributed by Kriton Kyrimis <kyrimis@cti.gr>)
  Greek characters for "wW" and "sS" are omitted to not produce a mismatch
  message since they have different characters in gr and el XFree86 layouts. */
static const char main_key_EL[MAIN_LEN][4] =
{
 "`~","1!","2@","3#","4$","5%","6^","7&","8*","9(","0)","-_","=+",
 "qQ;:","wW","eE","rR","tT","yY","uU","iI","oO","pP","[{","]}",
 "aA","sS","dD","fF","gG","hH","jJ","kK","lL",";:","'\"","\\|",
 "zZ","xX","cC","vV","bB","nN","mM",",<",".>","/?",
 "<>"
};

/*** Thai (Kedmanee) keyboard layout by Supphachoke Suntiwichaya <mrchoke@opentle.org> */
static const char main_key_th[MAIN_LEN][4] =
{
 "`~_%","1!+","2@/","3#-","4$","5%","6^","7&","8*","9(","0)","-_","=+",
 "qQ","wW\"","eEӮ","rR","tTи","yY","uU","iIó","oO","pP­","[{","]},",
 "aA","sS˦","dD","fF","gG","hH","jJ","kK","lL",";:ǫ","\'\".","\\|",
 "zZ(","xX)","cC","vV","bB","nN","mM?",",<",".>","/?"
}; 

/*** VNC keyboard layout */
static const WORD main_key_scan_vnc[MAIN_LEN] =
{
   0x02,0x03,0x04,0x05,0x06,0x07,0x08,0x09,0x0A,0x0B,0x0C,0x0D,0x1A,0x1B,0x27,0x28,0x29,0x33,0x34,0x35,0x2B,
   0x1E,0x30,0x2E,0x20,0x12,0x21,0x22,0x23,0x17,0x24,0x25,0x26,0x32,0x31,0x18,0x19,0x10,0x13,0x1F,0x14,0x16,0x2F,0x11,0x2D,0x15,0x2C,
   0x56
};

static const WORD main_key_vkey_vnc[MAIN_LEN] =
{
   '1','2','3','4','5','6','7','8','9','0',VK_OEM_MINUS,VK_OEM_PLUS,VK_OEM_4,VK_OEM_6,VK_OEM_1,VK_OEM_7,VK_OEM_3,VK_OEM_COMMA,VK_OEM_PERIOD,VK_OEM_2,VK_OEM_5,
   'A','B','C','D','E','F','G','H','I','J','K','L','M','N','O','P','Q','R','S','T','U','V','W','X','Y','Z',
   VK_OEM_102
};

static const char main_key_vnc[MAIN_LEN][4] =
{
 "1!","2@","3#","4$","5%","6^","7&","8*","9(","0)","-_","=+","[{","]}",";:","'\"","`~",",<",".>","/?","\\|",
 "aA","bB","cC","dD","eE","fF","gG","hH","iI","jJ","kK","lL","mM","nN","oO","pP","qQ","rR","sS","tT","uU","vV","wW","xX","yY","zZ"
};

/*** Dutch keyboard layout (setxkbmap nl) ***/
static const char main_key_NL[MAIN_LEN][4] =
{
 "@","1!","2\"","3#","4$","5%","6&","7_","8(","9)","0'","/?","~",
 "qQ","wW","eE","rR","tT","yY","uU","iI","oO","pP","~","*|",
 "aA","sS","dD","fF","gG","hH","jJ","kK","lL","+","'`","<>",
 "zZ","xX","cC","vV","bB","nN","mM",",;",".:","-=",
 "[]"
};



/*** Layout table. Add your keyboard mappings to this list */
static struct {
    const LCID lcid; /* input locale identifier, look for LOCALE_ILANGUAGE
                in the appropriate dlls/kernel/nls/.nls file */
    unsigned int codepage; /* If keysyms in a layout are outside of Your locale,
                but You want to input them, then Wine will try to simply cut out
                the lower byte of each keysym (raw lookup) instead of converting
                to current locale (locale lookup).
                  If chars, produced this way, match some Windows codepage, write
                it here.
                  0 = DefaultUnixCodepage for lcid (fallback), better to
                replace by the best candidate, if it exists and You know
                it exactly.
                */
    const char * const comment;
    const char (* const key)[MAIN_LEN][4];
    const WORD (* const scan)[MAIN_LEN]; /* scan codes mapping */
    const WORD (* const vkey)[MAIN_LEN]; /* virtual key codes mapping */
} main_key_tab[]={
 {0x0409,     0, "United States keyboard layout", &main_key_US, &main_key_scan_qwerty, &main_key_vkey_qwerty},
 {0x0409,     0, "United States keyboard layout (phantom key version)", &main_key_US_phantom, &main_key_scan_qwerty, &main_key_vkey_qwerty},
 {0x0409,     0, "United States keyboard layout (dvorak)", &main_key_US_dvorak, &main_key_scan_dvorak, &main_key_vkey_dvorak},
 {0x0409,     0, "United States International keyboard layout", &main_key_US_intl, &main_key_scan_qwerty, &main_key_vkey_qwerty},
 {0x0809,     0, "British keyboard layout", &main_key_UK, &main_key_scan_qwerty, &main_key_vkey_qwerty},
 {0x0407,     0, "German keyboard layout", &main_key_DE, &main_key_scan_qwerty, &main_key_vkey_qwertz},
 {0x0407,     0, "German keyboard layout without dead keys", &main_key_DE_nodead, &main_key_scan_qwerty, &main_key_vkey_qwertz},
 {0x0407,     0, "German keyboard layout for logitech desktop pro", &main_key_DE_logitech,  &main_key_scan_qwerty, &main_key_vkey_qwertz},
 {0x0407,     0, "German keyboard layout without dead keys 105", &main_key_DE_nodead_105, &main_key_scan_qwerty, &main_key_vkey_qwertz_105},
 {0x0407,     0, "German keyboard layout (simple)", &main_key_DE_std, &main_key_scan_qwerty, &main_key_vkey_qwerty},
 {0x0807,     0, "Swiss German keyboard layout", &main_key_SG, &main_key_scan_qwerty, &main_key_vkey_qwertz},
 {0x100c,     0, "Swiss French keyboard layout", &main_key_SF, &main_key_scan_qwerty, &main_key_vkey_qwertz},
 {0x041d,     0, "Swedish keyboard layout", &main_key_SE, &main_key_scan_qwerty, &main_key_vkey_qwerty_v2},
 {0x0425,     0, "Estonian keyboard layout", &main_key_ET, &main_key_scan_qwerty, &main_key_vkey_qwerty},
 {0x0414,     0, "Norwegian keyboard layout", &main_key_NO, &main_key_scan_qwerty, &main_key_vkey_qwerty},
 {0x0406,     0, "Danish keyboard layout", &main_key_DA, &main_key_scan_qwerty, &main_key_vkey_qwerty},
 {0x040c,     0, "French keyboard layout", &main_key_FR, &main_key_scan_qwerty, &main_key_vkey_azerty},
 {0x0c0c,     0, "Canadian French keyboard layout", &main_key_CF, &main_key_scan_qwerty, &main_key_vkey_qwerty},
 {0x0c0c,     0, "Canadian French keyboard layout (CA_fr)", &main_key_CA_fr, &main_key_scan_qwerty, &main_key_vkey_qwerty},
 {0x0c0c,     0, "Canadian keyboard layout", &main_key_CA, &main_key_scan_qwerty, &main_key_vkey_qwerty},
 {0x080c,     0, "Belgian keyboard layout", &main_key_BE, &main_key_scan_qwerty, &main_key_vkey_azerty},
 {0x0816,     0, "Portuguese keyboard layout", &main_key_PT, &main_key_scan_qwerty, &main_key_vkey_qwerty},
 {0x0416,     0, "Brazilian ABNT-2 keyboard layout", &main_key_PT_br, &main_key_scan_abnt_qwerty, &main_key_vkey_abnt_qwerty},
 {0x0416,     0, "Brazilian ABNT-2 keyboard layout ALT GR", &main_key_PT_br_alt_gr,&main_key_scan_abnt_qwerty, &main_key_vkey_abnt_qwerty},
 {0x040b,     0, "Finnish keyboard layout", &main_key_FI, &main_key_scan_qwerty, &main_key_vkey_qwerty},
 {0x0402,     0, "Bulgarian bds keyboard layout", &main_key_BG_bds, &main_key_scan_qwerty, &main_key_vkey_qwerty},
 {0x0402,     0, "Bulgarian phonetic keyboard layout", &main_key_BG_phonetic, &main_key_scan_qwerty, &main_key_vkey_qwerty},
 {0x0423, 21866, "Belarusian keyboard layout", &main_key_BY, &main_key_scan_qwerty, &main_key_vkey_qwerty},
 {0x0419, 20866, "Russian keyboard layout", &main_key_RU, &main_key_scan_qwerty, &main_key_vkey_qwerty},
 {0x0419, 20866, "Russian keyboard layout (phantom key version)", &main_key_RU_phantom, &main_key_scan_qwerty, &main_key_vkey_qwerty},
 {0x0419, 20866, "Russian keyboard layout KOI8-R", &main_key_RU_koi8r, &main_key_scan_qwerty, &main_key_vkey_qwerty},
 {0x0419,  1251, "Russian keyboard layout cp1251", &main_key_RU_cp1251, &main_key_scan_qwerty, &main_key_vkey_qwerty},
 {0x0419, 20866, "Russian phonetic keyboard layout", &main_key_RU_phonetic, &main_key_scan_qwerty, &main_key_vkey_qwerty},
 {0x0422, 21866, "Ukrainian keyboard layout KOI8-U", &main_key_UA, &main_key_scan_qwerty, &main_key_vkey_qwerty},
 {0x0422, 21866, "Ukrainian keyboard layout (standard)", &main_key_UA_std, &main_key_scan_qwerty, &main_key_vkey_qwerty},
 {0x0419, 20866, "Russian keyboard layout (standard)", &main_key_RU_std, &main_key_scan_qwerty, &main_key_vkey_qwerty},
 {0x0423, 21866, "Belarusian keyboard layout (standard)", &main_key_BY_std, &main_key_scan_qwerty, &main_key_vkey_qwerty},
 {0x040a,     0, "Spanish keyboard layout", &main_key_ES, &main_key_scan_qwerty, &main_key_vkey_qwerty},
 {0x0410,     0, "Italian keyboard layout", &main_key_IT, &main_key_scan_qwerty, &main_key_vkey_qwerty},
 {0x040f,     0, "Icelandic keyboard layout", &main_key_IS, &main_key_scan_qwerty, &main_key_vkey_qwerty},
 {0x040e,     0, "Hungarian keyboard layout", &main_key_HU, &main_key_scan_qwerty, &main_key_vkey_qwertz},
 {0x0415,     0, "Polish (programmer's) keyboard layout", &main_key_PL, &main_key_scan_qwerty, &main_key_vkey_qwerty},
 {0x0415,     0, "Polish keyboard layout (simple)", &main_key_PL_std, &main_key_scan_qwerty, &main_key_vkey_qwerty},
 {0x0424,     0, "Slovenian keyboard layout", &main_key_SI, &main_key_scan_qwerty, &main_key_vkey_qwertz},
 {0x0c1a,     0, "Serbian keyboard layout sr", &main_key_SR, &main_key_scan_qwerty, &main_key_vkey_qwerty}, /* LANG_SERBIAN,SUBLANG_SERBIAN_CYRILLIC */
 {0x0c1a,     0, "Serbian keyboard layout us,sr", &main_key_US_SR, &main_key_scan_qwerty, &main_key_vkey_qwerty}, /* LANG_SERBIAN,SUBLANG_SERBIAN_CYRILLIC */
 {0x041a,     0, "Croatian keyboard layout", &main_key_HR, &main_key_scan_qwerty, &main_key_vkey_qwertz},
 {0x041a,     0, "Croatian keyboard layout (specific)", &main_key_HR_jelly, &main_key_scan_qwerty, &main_key_vkey_qwerty},
 {0x0411,     0, "Japanese 106 keyboard layout", &main_key_JA_jp106, &main_key_scan_qwerty_jp106, &main_key_vkey_qwerty_jp106},
 {0x0411,     0, "Japanese pc98x1 keyboard layout", &main_key_JA_pc98x1, &main_key_scan_qwerty, &main_key_vkey_qwerty},
 {0x041b,     0, "Slovak keyboard layout", &main_key_SK, &main_key_scan_qwerty, &main_key_vkey_qwerty},
 {0x041b,     0, "Slovak and Czech keyboard layout without dead keys", &main_key_SK_prog, &main_key_scan_qwerty, &main_key_vkey_qwerty},
 {0x0405,     0, "Czech keyboard layout", &main_key_CS, &main_key_scan_qwerty, &main_key_vkey_qwerty},
 {0x0405,     0, "Czech keyboard layout cz", &main_key_CZ, &main_key_scan_qwerty, &main_key_vkey_qwertz},
 {0x0405,     0, "Czech keyboard layout cz_qwerty", &main_key_CZ_qwerty, &main_key_scan_qwerty, &main_key_vkey_qwerty},
 {0x040a,     0, "Latin American keyboard layout", &main_key_LA, &main_key_scan_qwerty, &main_key_vkey_qwerty},
 {0x0427,     0, "Lithuanian (Baltic) keyboard layout", &main_key_LT_B, &main_key_scan_qwerty, &main_key_vkey_qwerty},
 {0x041f,     0, "Turkish keyboard layout", &main_key_TK, &main_key_scan_qwerty, &main_key_vkey_qwerty},
 {0x041f,     0, "Turkish keyboard layout tr", &main_key_TR, &main_key_scan_qwerty, &main_key_vkey_qwerty},
 {0x041f,     0, "Turkish keyboard layout trf", &main_key_TR_F, &main_key_scan_qwerty, &main_key_vkey_qwerty},
 {0x040d,     0, "Israelian keyboard layout", &main_key_IL, &main_key_scan_qwerty, &main_key_vkey_qwerty},
 {0x040d,     0, "Israelian phonetic keyboard layout", &main_key_IL_phonetic, &main_key_scan_qwerty, &main_key_vkey_qwerty},
 {0x040d,     0, "Israelian Saharon keyboard layout", &main_key_IL_saharon, &main_key_scan_qwerty, &main_key_vkey_qwerty},
 {0x0409,     0, "VNC keyboard layout", &main_key_vnc, &main_key_scan_vnc, &main_key_vkey_vnc},
 {0x0408,     0, "Greek keyboard layout", &main_key_EL, &main_key_scan_qwerty, &main_key_vkey_qwerty},
 {0x041e,     0, "Thai (Kedmanee)  keyboard layout", &main_key_th, &main_key_scan_qwerty, &main_key_vkey_qwerty},
 {0x0413,     0, "Dutch keyboard layout", &main_key_NL, &main_key_scan_qwerty, &main_key_vkey_qwerty},

 {0, 0, NULL, NULL, NULL, NULL} /* sentinel */
};

/* maybe more of these scancodes should be extended? */
                /* extended must be set for ALT_R, CTRL_R,
                   INS, DEL, HOME, END, PAGE_UP, PAGE_DOWN, ARROW keys,
                   keypad / and keypad ENTER (SDK 3.1 Vol.3 p 138) */
                /* FIXME should we set extended bit for NumLock ? My
                 * Windows does ... DF */
                /* Yes, to distinguish based on scan codes, also
                   for PrtScn key ... GA */

static const WORD nonchar_key_vkey[256] =
{
    /* unused */
    0, 0, 0, 0, 0, 0, 0, 0,                                     /* FF00 */
    /* special keys */
    VK_BACK, VK_TAB, 0, VK_CLEAR, 0, VK_RETURN, 0, 0,           /* FF08 */
    0, 0, 0, VK_PAUSE, VK_SCROLL, 0, 0, 0,                      /* FF10 */
    0, 0, 0, VK_ESCAPE, 0, 0, 0, 0,                             /* FF18 */
    /* unused */
    0, 0, 0, 0, 0, 0, 0, 0,                                     /* FF20 */
    0, 0, 0, 0, 0, 0, 0, 0,                                     /* FF28 */
    0, 0, 0, 0, 0, 0, 0, 0,                                     /* FF30 */
    0, 0, 0, 0, 0, 0, 0, 0,                                     /* FF38 */
    0, 0, 0, 0, 0, 0, 0, 0,                                     /* FF40 */
    0, 0, 0, 0, 0, 0, 0, 0,                                     /* FF48 */
    /* cursor keys */
    VK_HOME, VK_LEFT, VK_UP, VK_RIGHT,                          /* FF50 */
    VK_DOWN, VK_PRIOR, VK_NEXT, VK_END,
    0, 0, 0, 0, 0, 0, 0, 0,                                     /* FF58 */
    /* misc keys */
    VK_SELECT, VK_SNAPSHOT, VK_EXECUTE, VK_INSERT, 0,0,0, VK_APPS, /* FF60 */
    0, VK_CANCEL, VK_HELP, VK_CANCEL, 0, 0, 0, 0,               /* FF68 */
    0, 0, 0, 0, 0, 0, 0, 0,                                     /* FF70 */
    /* keypad keys */
    0, 0, 0, 0, 0, 0, 0, VK_NUMLOCK,                            /* FF78 */
    0, 0, 0, 0, 0, 0, 0, 0,                                     /* FF80 */
    0, 0, 0, 0, 0, VK_RETURN, 0, 0,                             /* FF88 */
    0, 0, 0, 0, 0, VK_HOME, VK_LEFT, VK_UP,                     /* FF90 */
    VK_RIGHT, VK_DOWN, VK_PRIOR, VK_NEXT,                       /* FF98 */
    VK_END, VK_CLEAR, VK_INSERT, VK_DELETE,
    0, 0, 0, 0, 0, 0, 0, 0,                                     /* FFA0 */
    0, 0, VK_MULTIPLY, VK_ADD,                                  /* FFA8 */
    VK_SEPARATOR, VK_SUBTRACT, VK_DECIMAL, VK_DIVIDE,
    VK_NUMPAD0, VK_NUMPAD1, VK_NUMPAD2, VK_NUMPAD3,             /* FFB0 */
    VK_NUMPAD4, VK_NUMPAD5, VK_NUMPAD6, VK_NUMPAD7,
    VK_NUMPAD8, VK_NUMPAD9, 0, 0, 0, VK_OEM_NEC_EQUAL,          /* FFB8 */
    /* function keys */
    VK_F1, VK_F2,
    VK_F3, VK_F4, VK_F5, VK_F6, VK_F7, VK_F8, VK_F9, VK_F10,    /* FFC0 */
    VK_F11, VK_F12, VK_F13, VK_F14, VK_F15, VK_F16, 0, 0,       /* FFC8 */
    0, 0, 0, 0, 0, 0, 0, 0,                                     /* FFD0 */
    0, 0, 0, 0, 0, 0, 0, 0,                                     /* FFD8 */
    /* modifier keys */
    0, VK_SHIFT, VK_SHIFT, VK_CONTROL,                          /* FFE0 */
    VK_CONTROL, VK_CAPITAL, 0, VK_MENU,
    VK_MENU, VK_MENU, VK_MENU, 0, 0, 0, 0, 0,                   /* FFE8 */
    0, 0, 0, 0, 0, 0, 0, 0,                                     /* FFF0 */
    0, 0, 0, 0, 0, 0, 0, VK_DELETE                              /* FFF8 */
};

static const WORD nonchar_key_scan[256] =
{
    /* unused */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,              /* FF00 */
    /* special keys */
    0x0E, 0x0F, 0x00, /*?*/ 0, 0x00, 0x1C, 0x00, 0x00,           /* FF08 */
    0x00, 0x00, 0x00, 0x45, 0x46, 0x00, 0x00, 0x00,              /* FF10 */
    0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00, 0x00,              /* FF18 */
    /* unused */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,              /* FF20 */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,              /* FF28 */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,              /* FF30 */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,              /* FF38 */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,              /* FF40 */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,              /* FF48 */
    /* cursor keys */
    0x147, 0x14B, 0x148, 0x14D, 0x150, 0x149, 0x151, 0x14F,      /* FF50 */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,              /* FF58 */
    /* misc keys */
    /*?*/ 0, 0x137, /*?*/ 0, 0x152, 0x00, 0x00, 0x00, 0x00,      /* FF60 */
    /*?*/ 0, /*?*/ 0, 0x38, 0x146, 0x00, 0x00, 0x00, 0x00,       /* FF68 */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,              /* FF70 */
    /* keypad keys */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x138, 0x145,            /* FF78 */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,              /* FF80 */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x11C, 0x00, 0x00,             /* FF88 */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x47, 0x4B, 0x48,              /* FF90 */
    0x4D, 0x50, 0x49, 0x51, 0x4F, 0x4C, 0x52, 0x53,              /* FF98 */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,              /* FFA0 */
    0x00, 0x00, 0x37, 0x4E, 0x53, 0x4A, 0x53, 0x135,             /* FFA8 */
    0x52, 0x4F, 0x50, 0x51, 0x4B, 0x4C, 0x4D, 0x47,              /* FFB0 */
    0x48, 0x49, 0x00, 0x00, 0x00, 0x00,                          /* FFB8 */
    /* function keys */
    0x3B, 0x3C,
    0x3D, 0x3E, 0x3F, 0x40, 0x41, 0x42, 0x43, 0x44,              /* FFC0 */
    0x57, 0x58, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,              /* FFC8 */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,              /* FFD0 */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,              /* FFD8 */
    /* modifier keys */
    0x00, 0x2A, 0x36, 0x1D, 0x11D, 0x3A, 0x00, 0x38,             /* FFE0 */
    0x138, 0x38, 0x138, 0x00, 0x00, 0x00, 0x00, 0x00,            /* FFE8 */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,              /* FFF0 */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x153              /* FFF8 */
};

static const WORD xfree86_vendor_key_vkey[256] =
{
    0, 0, 0, 0, 0, 0, 0, 0,                                     /* 1008FF00 */
    0, 0, 0, 0, 0, 0, 0, 0,                                     /* 1008FF08 */
    0, VK_VOLUME_DOWN, VK_VOLUME_MUTE, VK_VOLUME_UP,            /* 1008FF10 */
    VK_MEDIA_PLAY_PAUSE, VK_MEDIA_STOP,
    VK_MEDIA_PREV_TRACK, VK_MEDIA_NEXT_TRACK,
    0, VK_LAUNCH_MAIL, 0, VK_BROWSER_SEARCH,                    /* 1008FF18 */
    0, 0, 0, VK_BROWSER_HOME,
    0, 0, 0, 0, 0, 0, VK_BROWSER_BACK, VK_BROWSER_FORWARD,      /* 1008FF20 */
    VK_BROWSER_STOP, VK_BROWSER_REFRESH, 0, 0, 0, 0, 0, 0,      /* 1008FF28 */
    VK_BROWSER_FAVORITES, 0, VK_LAUNCH_MEDIA_SELECT, 0,         /* 1008FF30 */
    0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0,                                     /* 1008FF38 */
    0, 0, 0, 0, 0, 0, 0, 0,                                     /* 1008FF40 */
    0, 0, 0, 0, 0, 0, 0, 0,                                     /* 1008FF48 */
    0, 0, 0, 0, 0, 0, 0, 0,                                     /* 1008FF50 */
    0, 0, 0, 0, 0, 0, 0, 0,                                     /* 1008FF58 */
    0, 0, 0, 0, 0, 0, 0, 0,                                     /* 1008FF60 */
    0, 0, 0, 0, 0, 0, 0, 0,                                     /* 1008FF68 */
    0, 0, 0, 0, 0, 0, 0, 0,                                     /* 1008FF70 */
    0, 0, 0, 0, 0, 0, 0, 0,                                     /* 1008FF78 */
    0, 0, 0, 0, 0, 0, 0, 0,                                     /* 1008FF80 */
    0, 0, 0, 0, 0, 0, 0, 0,                                     /* 1008FF88 */
    0, 0, 0, 0, 0, 0, 0, 0,                                     /* 1008FF90 */
    0, 0, 0, 0, 0, 0, 0, 0,                                     /* 1008FF98 */
    0, 0, 0, 0, 0, 0, 0, 0,                                     /* 1008FFA0 */
    0, 0, 0, 0, 0, 0, 0, 0,                                     /* 1008FFA8 */
    0, 0, 0, 0, 0, 0, 0, 0,                                     /* 1008FFB0 */
    0, 0, 0, 0, 0, 0, 0, 0,                                     /* 1008FFB8 */
    0, 0, 0, 0, 0, 0, 0, 0,                                     /* 1008FFC0 */
    0, 0, 0, 0, 0, 0, 0, 0,                                     /* 1008FFC8 */
    0, 0, 0, 0, 0, 0, 0, 0,                                     /* 1008FFD0 */
    0, 0, 0, 0, 0, 0, 0, 0,                                     /* 1008FFD8 */
    0, 0, 0, 0, 0, 0, 0, 0,                                     /* 1008FFE0 */
    0, 0, 0, 0, 0, 0, 0, 0,                                     /* 1008FFE8 */
    0, 0, 0, 0, 0, 0, 0, 0,                                     /* 1008FFF0 */
    0, 0, 0, 0, 0, 0, 0, 0                                      /* 1008FFF8 */
};

/* Forward declarations */

static void
KEYBOARD_DetectLayouts(void);

void
X11DRV_InitKeyboard(void);

inline static KeyCode
OverToKeyc(KeyCode k);

HKL X11DRV_GetKeyboardLayout(DWORD);

/**********************************************************************
 *		KEYBOARD_LayoutSwitched
 *
 *     Track changes of effective XKB group and overlay group
 */

static Bool
KEYBOARD_LayoutSwitched (void)
{ /* enter and exit in tsx11_lock'ed state */
  int changed = 0;
  Display *display = thread_display();
#ifdef HAVE_XKB
  if (use_xkb) {
    if (!(desc = XkbAllocKeyboard())) {
       use_xkb = 0;
       ERR("Cannot allocate memory for XKB controls, XKB disabled!\n");
    }
  }
  if (use_xkb) {
    unsigned ctrls;
    int min_key_code, max_key_code;
    XkbStateRec state;
    XkbGetState(display, XkbUseCoreKbd, &state);
    kbd.keysyms_per_keycode = 4; /* ??? */
    kbd.group = state.locked_group;
    XkbGetControls(display, XkbAllControlsMask, desc);
    ctrls = desc->ctrls->enabled_ctrls;
    kbd.overlay = 0;
    if (ctrls & XkbOverlay1Mask) kbd.overlay  = OVER1_ENABLED;
    if (ctrls & XkbOverlay2Mask) kbd.overlay |= OVER2_ENABLED;
    kbd.overlay &= kbd.overlays_used;
    changed = MAX_OVERLAYS*kbd.group + kbd.overlay;
    if (kbd.layout_no != changed) {
        kbd.layout_no = changed;
        changed = 1;
    } else changed = 0;
    XDisplayKeycodes(display,&min_key_code,&max_key_code);
    if (min_key_code != kbd.min_keycode) {
        TRACE(": min_key_code changed from %d to %d\n",
          kbd.min_keycode, min_key_code); 
        kbd.min_keycode = min_key_code;
        changed = global_refresh = 1;
    }
    if (max_key_code != kbd.max_keycode) {
        TRACE(": max_key_code changed from %d to %d\n",
          kbd.max_keycode, max_key_code); 
        kbd.max_keycode = max_key_code;
        changed = global_refresh = 1;
    }
    if (desc->ctrls->num_groups != kbd.num_groups) {
        TRACE(": num_groups changed from %d to %d\n",
          kbd.num_groups, desc->ctrls->num_groups); 
        kbd.num_groups = desc->ctrls->num_groups;
        changed = global_refresh = 1;
    }
    if (global_refresh)
      TRACE("caused global refresh: %d groups, keycodes from %d to %d\n",
          kbd.num_groups, kbd.min_keycode, kbd.max_keycode); 
    else
      TRACE("Selected XKB group %d, overlay %d\n",
          kbd.group, kbd.overlay);
  } else
#endif /* HAVE_XKB */
  {
    int min_key_code, max_key_code, have_chars;
    kbd.layout_no = 0;
    kbd.num_groups = kbd.num_layouts = 1;
    changed   = 0;
    XDisplayKeycodes(display, &min_key_code, &max_key_code);
    XFree(XGetKeyboardMapping(display,
      min_key_code, max_key_code - min_key_code +1, &have_chars));
    if ((min_key_code != kbd.min_keycode) || (max_key_code != kbd.max_keycode) ||
        (have_chars != kbd.keysyms_per_keycode)) {
        kbd.min_keycode = min_key_code;
        kbd.max_keycode = max_key_code;
        kbd.keysyms_per_keycode = have_chars;
        changed = global_refresh = 1;
        TRACE("caused global refresh: max %d keysyms, keycodes from %d to %d\n",
          have_chars, kbd.min_keycode, kbd.max_keycode);
    }
  }
#ifdef HAVE_XKB
  if (use_xkb && desc) {
      XkbFreeKeyboard(desc, XkbControlsMask,True);
      desc = NULL;
  }
  if (changed) {
    if (global_refresh || (kbd.layout[kbd.layout_no] == -1)) 
     TRACE("to layout #%d - detect it later\n",kbd.layout_no);
    else {
     HWND hwnd = GetFocus();
     if (!hwnd) hwnd = GetActiveWindow();
     PostMessageW(hwnd, WM_INPUTLANGCHANGEREQUEST,
               0 /*FIXME*/, (LPARAM)X11DRV_GetKeyboardLayout(0));
     TRACE("to layout #%d, \"%s\", codepage %d\n",
         kbd.layout_no, main_key_tab[kbd.layout[kbd.layout_no]].comment,kbd.cp[kbd.layout_no]);
    }
  }
#endif
  return changed;
}

/**********************************************************************
 *		EVENT_event_to_vkey
 *
 */

/* Returns the Windows virtual key code associated with the X event <e> */
/* x11 lock must be held */
static WORD EVENT_event_to_vkey( XIC xic, XKeyEvent *e)
{
    KeyCode kc = 0;
    KeySym keysym = 0;
    Status status;
    char buf[24];
    WORD vkey = 0;
    int savetype, check, i;
    unsigned int savestate;

    TRACE_(key)("Received e->keycode = %x\n", e->keycode);

    savestate = e->state;
    if (xic) {
         /* XmbLookupString returns neither char nor keysym for KeyRelease */
          savetype = e->type;
          if (savetype == KeyRelease) e->type = KeyPress;
          XmbLookupString(xic, e, buf, sizeof(buf), &keysym, &status);
          e->type = savetype;
        }
    else
        XLookupString(e, buf, sizeof(buf), &keysym, NULL);
    e->state = savestate;


#ifdef HAVE_XKB
    if (use_xkb) check = 2;
    else
#endif
    check = 1;

  again:
    kc = OverToKeyc(e->keycode);
    if (kc != e->keycode)
       TRACE_(key)("Treated as e->keycode = %x\n", kc);

    if ((e->state & NumLockMask) &&
        (keysym == XK_KP_Separator || keysym == XK_KP_Decimal ||
         (keysym >= XK_KP_0 && keysym <= XK_KP_9)))
        /* Only the Keypad keys 0-9 and . send different keysyms
         * depending on the NumLock state */
        vkey = nonchar_key_vkey[keysym & 0xFF];
    else
        vkey = kbd.keyc[kbd.layout_no]->vkey[kc];

    for (i = 0; i < 4; i++) {
#ifdef HAVE_XKB
       if (use_xkb) {
         TRACE_(key)("Compare: 0x%04lx <> 0x%04lx, group %d, overlay %d, level %d\n",
             keysym,kbd.keyc[kbd.layout_no]->ksym[kc][i], kbd.group, kbd.overlay, i);
       } else
#endif
         TRACE_(key)("Compare: 0x%04lx <> 0x%04lx, level %d\n",
             keysym,kbd.keyc[kbd.layout_no]->ksym[kc][i], i);
       if (kbd.keyc[kbd.layout_no]->ksym[kc][i] && (keysym == kbd.keyc[kbd.layout_no]->ksym[kc][i])) {
            if (vkey) check = 0;
            break;
       }
    }

    switch (check--) {
         case 2:
           KEYBOARD_LayoutSwitched();
           if (global_refresh) {
               TRACE_(key)("Redetection initiated\n");
               /* to avoid lock() inside lock() */
               wine_tsx11_unlock();
               KEYBOARD_DetectLayouts();
               wine_tsx11_lock();
           }
           goto again;
           case 1:
         TRACE_(key)("Matching keycode and level not found for keysym 0x%04lx in layout #%d\n",
             keysym,kbd.layout_no);
    }

    e->keycode = kc;
    return vkey;
}



static BOOL NumState=FALSE, CapsState=FALSE;


/***********************************************************************
 *		X11DRV_send_keyboard_input
 */
void X11DRV_send_keyboard_input( WORD wVk, WORD wScan, DWORD dwFlags, DWORD time,
                                 DWORD dwExtraInfo, UINT injected_flags )
{
    UINT message;
    KEYLP keylp;
    KBDLLHOOKSTRUCT hook;
    WORD wVkStripped;

    wVk = LOBYTE(wVk);

    /* strip left/right for menu, control, shift */
    if (wVk == VK_LMENU || wVk == VK_RMENU)
        wVkStripped = VK_MENU;
    else if (wVk == VK_LCONTROL || wVk == VK_RCONTROL)
        wVkStripped = VK_CONTROL;
    else if (wVk == VK_LSHIFT || wVk == VK_RSHIFT)
        wVkStripped = VK_SHIFT;
    else
        wVkStripped = wVk;

    keylp.lp2 = 0;
    keylp.lp1.count = 1;
    keylp.lp1.code = wScan;
    keylp.lp1.extended = (dwFlags & KEYEVENTF_EXTENDEDKEY) != 0;
    keylp.lp1.win_internal = 0; /* this has something to do with dialogs,
                                * don't remember where I read it - AK */
                                /* it's '1' under windows, when a dialog box appears
                                 * and you press one of the underlined keys - DF*/

    /* note that there is a test for all this */
    if (dwFlags & KEYEVENTF_KEYUP )
    {
        message = WM_KEYUP;
        if ((key_state_table[VK_MENU] & 0x80) &&
            ((wVkStripped == VK_MENU) || (wVkStripped == VK_CONTROL)
             || !(key_state_table[VK_CONTROL] & 0x80)))
        {
            if( TrackSysKey == VK_MENU || /* <ALT>-down/<ALT>-up sequence */
                (wVkStripped != VK_MENU)) /* <ALT>-down...<something else>-up */
                message = WM_SYSKEYUP;
            TrackSysKey = 0;
        }
        key_state_table[wVk] &= ~0x80;
        key_state_table[wVkStripped] &= ~0x80;
        keylp.lp1.previous = 1;
        keylp.lp1.transition = 1;
    }
    else
    {
        keylp.lp1.previous = (key_state_table[wVk] & 0x80) != 0;
        keylp.lp1.transition = 0;
        if (!(key_state_table[wVk] & 0x80)) key_state_table[wVk] ^= 0x01;
        key_state_table[wVk] |= 0xc0;
        key_state_table[wVkStripped] |= 0xc0;

        message = WM_KEYDOWN;
        if ((key_state_table[VK_MENU] & 0x80) && !(key_state_table[VK_CONTROL] & 0x80))
        {
            message = WM_SYSKEYDOWN;
            TrackSysKey = wVkStripped;
        }
    }

    keylp.lp1.context = (key_state_table[VK_MENU] & 0x80) != 0; /* 1 if alt */

    TRACE_(key)(" wParam=%04x, lParam=%08lx, InputKeyState=%x\n",
                wVk, keylp.lp2, key_state_table[wVk] );

    hook.vkCode      = wVk;
    hook.scanCode    = wScan;
    hook.flags       = (keylp.lp2 >> 24) | injected_flags;
    hook.time        = time;
    hook.dwExtraInfo = dwExtraInfo;
    if (HOOK_CallHooks( WH_KEYBOARD_LL, HC_ACTION, message, (LPARAM)&hook, TRUE )) return;

    SERVER_START_REQ( send_hardware_message )
    {
        req->id       = (injected_flags & LLKHF_INJECTED) ? 0 : GetCurrentThreadId();
        req->win      = 0;
        req->msg      = message;
        req->wparam   = wVk;
        req->lparam   = keylp.lp2;
        req->x        = cursor_pos.x;
        req->y        = cursor_pos.y;
        req->time     = time;
        req->info     = dwExtraInfo;
        wine_server_call( req );
    }
    SERVER_END_REQ;
}


/**********************************************************************
 *		KEYBOARD_GenerateMsg
 *
 * Generate Down+Up messages when NumLock or CapsLock is pressed.
 *
 * Convention : called with vkey only VK_NUMLOCK or VK_CAPITAL
 *
 */
static void KEYBOARD_GenerateMsg( WORD vkey, WORD scan, int Evtype, DWORD event_time )
{
  BOOL * State = (vkey==VK_NUMLOCK? &NumState : &CapsState);
  DWORD up, down;

  if (*State) {
    /* The INTERMEDIARY state means : just after a 'press' event, if a 'release' event comes,
       don't treat it. It's from the same key press. Then the state goes to ON.
       And from there, a 'release' event will switch off the toggle key. */
    *State=FALSE;
    TRACE_(key)("INTERM : don't treat release of toggle key. key_state_table[%#x] = %#x\n",
          vkey,key_state_table[vkey]);
  } else
    {
        down = (vkey==VK_NUMLOCK ? KEYEVENTF_EXTENDEDKEY : 0);
        up = (vkey==VK_NUMLOCK ? KEYEVENTF_EXTENDEDKEY : 0) | KEYEVENTF_KEYUP;
        if ( key_state_table[vkey] & 0x1 ) /* it was ON */
          {
            if (Evtype!=KeyPress)
              {
                TRACE_(key)("ON + KeyRelease => generating DOWN and UP messages.\n");
                X11DRV_send_keyboard_input( vkey, scan, down, event_time, 0, 0 );
                X11DRV_send_keyboard_input( vkey, scan, up, event_time, 0, 0 );
                *State=FALSE;
                key_state_table[vkey] &= ~0x01; /* Toggle state to off. */
              }
          }
        else /* it was OFF */
          if (Evtype==KeyPress)
            {
              TRACE_(key)("OFF + Keypress => generating DOWN and UP messages.\n");
              X11DRV_send_keyboard_input( vkey, scan, down, event_time, 0, 0 );
              X11DRV_send_keyboard_input( vkey, scan, up, event_time, 0, 0 );
              *State=TRUE; /* Goes to intermediary state before going to ON */
              key_state_table[vkey] |= 0x01; /* Toggle state to on. */
            }
    }
}

/***********************************************************************
 *		KEYBOARD_UpdateOneState
 *
 * Updates internal state for <vkey>, depending on key <state> under X
 *
 */
inline static void KEYBOARD_UpdateOneState ( int vkey, int state, DWORD time )
{
    /* Do something if internal table state != X state for keycode */
    if (((key_state_table[vkey] & 0x80)!=0) != state)
    {
        TRACE("Adjusting state for vkey %#.2x. State before %#.2x\n",
              vkey, key_state_table[vkey]);

        /* Fake key being pressed inside wine */
        X11DRV_send_keyboard_input( vkey, 0, state? 0 : KEYEVENTF_KEYUP, time, 0, 0 );

        TRACE("State after %#.2x\n",key_state_table[vkey]);
    }
}

/***********************************************************************
 *		X11DRV_KeymapNotify
 *
 * Update modifiers state (Ctrl, Alt, Shift) when window is activated.
 *
 * This handles the case where one uses Ctrl+... Alt+... or Shift+.. to switch
 * from wine to another application and back.
 * Toggle keys are handled in HandleEvent.
 */
void X11DRV_KeymapNotify( HWND hwnd, XEvent *event )
{
    int i, j, alt, control, shift;
    DWORD time = GetCurrentTime();

    alt = control = shift = 0;
    /* the minimum keycode is always greater or equal to 8, so we can
     * skip the first 8 values, hence start at 1
     */
    for (i = 1; i < 32; i++)
    {
        if (!event->xkeymap.key_vector[i]) continue;
        for (j = 0; j < 8; j++)
        {
            if (!(event->xkeymap.key_vector[i] & (1<<j))) continue;
            switch(kbd.keyc[kbd.layout_no]->vkey[(i * 8) + j] & 0xff)
            {
            case VK_MENU:    alt = 1; break;
            case VK_CONTROL: control = 1; break;
            case VK_SHIFT:   shift = 1; break;
            }
        }
    }
    KEYBOARD_UpdateOneState( VK_MENU, alt, time );
    KEYBOARD_UpdateOneState( VK_CONTROL, control, time );
    KEYBOARD_UpdateOneState( VK_SHIFT, shift, time );
}

/***********************************************************************
 *		X11DRV_KeyEvent
 *
 *         Handle a X key event
 */
void X11DRV_KeyEvent( HWND hwnd, XEvent *xev )
{
    XKeyEvent *event = &xev->xkey;
    char Str[24];
    KeySym keysym = 0;
    WORD vkey = 0, bScan;
    DWORD dwFlags;
    int ascii_chars, savetype;
    unsigned int savestate;
    XIC xic = X11DRV_get_ic( hwnd );
    DWORD event_time = EVENT_x11_time_to_win32_time(event->time);
    Status status = 0;

    TRACE_(key)("type %d, window %lx, state 0x%04x, keycode 0x%04x\n",
                event->type, event->window, event->state, event->keycode);

    wine_tsx11_lock();

    savestate = event->state;
    if (xic) {
         /* XmbLookupString returns neither char nor keysym for KeyRelease */
          savetype = event->type;
          if (savetype == KeyRelease) event->type = KeyPress;
          ascii_chars = XmbLookupString(xic, event, Str, sizeof(Str), &keysym, &status);
          event->type = savetype;
        }
    else
        ascii_chars = XLookupString(event, Str, sizeof(Str), &keysym, NULL);
    event->state = savestate;

    wine_tsx11_unlock();

    /* FIXME: Ignore some unwanted events - think about it again */
    if ((keysym >= XK_ISO_Lock && keysym <= XK_ISO_Last_Group_Lock) ||
         keysym == XK_Overlay1_Enable || keysym == XK_Overlay2_Enable ||
         keysym == XK_Mode_switch)
    {
        wine_tsx11_lock();
        TRACE_(key)("Ignoring %s keyboard event\n", XKeysymToString(keysym));
        wine_tsx11_unlock();
        return;
    }

    TRACE_(key)("state = %X nbyte = %d, status 0x%x\n", event->state, ascii_chars, status);

    if (status == XBufferOverflow)
        ERR("Buffer Overflow need %i!\n",ascii_chars);

    if (status == XLookupChars)
    {
        X11DRV_XIMLookupChars( Str, ascii_chars );
        return;
    }

    /* If XKB extensions are used, the state mask for AltGr will use the group
       index instead of the modifier mask. The group index is set in bits
       13-14 of the state field in the XKeyEvent structure. So if AltGr is
       pressed, look if the group index is different than 0. From XKB
       extension documentation, the group index for AltGr should be 2
       (event->state = 0x2000). It's probably better to not assume a
       predefined group index and find it dynamically

       Ref: X Keyboard Extension: Library specification (section 14.1.1 and 17.1.1) */
    /* Save also all possible modifier states. */
    AltGrMask = event->state & (0x6000 | Mod1Mask | Mod2Mask | Mod3Mask | Mod4Mask | Mod5Mask);

    Str[ascii_chars] = '\0';
    if (TRACE_ON(key)){
        const char *ksname;

        wine_tsx11_lock();
        ksname = XKeysymToString(keysym);
        wine_tsx11_unlock();
        if (!ksname)
          ksname = "No Name";
        TRACE_(key)("%s : keysym=%lX (%s), # of chars=%d / 0x%02x / '%s'\n",
                    (event->type == KeyPress) ? "KeyPress" : "KeyRelease",
                    keysym, ksname, ascii_chars, Str[0] & 0xff, Str);
    }

    wine_tsx11_lock();
    vkey = EVENT_event_to_vkey(xic,event);
    /* X returns keycode 0 for composed characters */
    if (!vkey && ascii_chars) vkey = VK_NONAME;
    wine_tsx11_unlock();

    TRACE_(key)("keycode 0x%x converted to vkey 0x%x\n",
                event->keycode, vkey);

   if (vkey)
   {
    switch (vkey & 0xff)
    {
    case VK_NUMLOCK:
      KEYBOARD_GenerateMsg( VK_NUMLOCK, 0x45, event->type, event_time );
      break;
    case VK_CAPITAL:
      TRACE_(key)("Caps Lock event. (type %d). State before : %#.2x\n",event->type,key_state_table[vkey]);
      KEYBOARD_GenerateMsg( VK_CAPITAL, 0x3A, event->type, event_time );
      TRACE_(key)("State after : %#.2x\n",key_state_table[vkey]);
      break;
    default:
        /* Adjust the NUMLOCK state if it has been changed outside wine */
        if (!(key_state_table[VK_NUMLOCK] & 0x01) != !(event->state & NumLockMask))
          {
            TRACE_(key)("Adjusting NumLock state.\n");
            KEYBOARD_GenerateMsg( VK_NUMLOCK, 0x45, KeyPress, event_time );
            KEYBOARD_GenerateMsg( VK_NUMLOCK, 0x45, KeyRelease, event_time );
          }
        /* Adjust the CAPSLOCK state if it has been changed outside wine */
        if (!(key_state_table[VK_CAPITAL] & 0x01) != !(event->state & LockMask))
          {
            TRACE_(key)("Adjusting Caps Lock state.\n");
            KEYBOARD_GenerateMsg( VK_CAPITAL, 0x3A, KeyPress, event_time );
            KEYBOARD_GenerateMsg( VK_CAPITAL, 0x3A, KeyRelease, event_time );
          }
        /* Not Num nor Caps : end of intermediary states for both. */
        NumState = FALSE;
        CapsState = FALSE;

        bScan = kbd.keyc[kbd.layout_no]->scan[event->keycode] & 0xFF;
        TRACE_(key)("bScan = 0x%02x.\n", bScan);

        dwFlags = 0;
        if ( event->type == KeyRelease ) dwFlags |= KEYEVENTF_KEYUP;
        if ( vkey & 0x100 )              dwFlags |= KEYEVENTF_EXTENDEDKEY;

        X11DRV_send_keyboard_input( vkey & 0xff, bScan, dwFlags, event_time, 0, 0 );
    }
   }
}

#ifdef HAVE_XKB

static const char * const xkb_event_names[] =
{
  "XkbNewKeyboardNotify", "XkbMapNotify", "XkbStateNotify",
  "XkbControlsNotify", "XkbIndicatorStateNotify", "XkbIndicatorMapNotify",
  "XkbNamesNotify", "XkbCompatMapNotify", "XkbBellNotify", "XkbActionMessage",
  "XkbAccessXnotify", "XkbExtensionDeviceNotify"
};

/***********************************************************************
 *		X11DRV_XkbEvent
 *
 *         Handle an XKB event
 */
void X11DRV_XkbEvent( HWND hwnd, XEvent * event )
{
    XkbEvent* xkbevent = (XkbEvent*) event;
    TRACE_(key)(" : type %s\n",xkb_event_names[xkbevent->any.xkb_type]);
    switch ( xkbevent->any.xkb_type ) {
      case XkbMapNotify:
        global_refresh = 1;
      case XkbStateNotify:
        if ( xkbevent->any.xkb_type == XkbStateNotify ){
          TRACE_(key)(" : changed %x\n", xkbevent->state.changed);
          if ( !(xkbevent->state.changed &
             ( XkbGroupStateMask | XkbGroupBaseMask |
               XkbGroupLatchMask | XkbGroupLockMask ))) {
              TRACE_(key)(" : ignore XkbStateNotify events other than group changes\n");
              break;
          }
        }
      case XkbControlsNotify:
        wine_tsx11_lock();
        KEYBOARD_LayoutSwitched();
        wine_tsx11_unlock();
        if (global_refresh) KEYBOARD_DetectLayouts();
        break;
      case XkbNewKeyboardNotify:
        X11DRV_InitKeyboard(); break;
      default: 
        break;
    }
    
}

/**********************************************************************
 *		KEYBOARD_RefreshOverlays
 *
 *     Build the tables of conversions done by overlays that
 *     map some keys to other, so that the same key can behave
 *     differently in the same XKB group depending on active
 *     overlay group (e.g. Ukrainian "i" becomes Russian "yeru").
 *     This way You can join close enough layouts (e.g. Czech
 *     and Slovak) into a single group. Wine will detect them all.
 */

static void
KEYBOARD_RefreshOverlays(Display *display) {
  KeyCode keyc, over;
  int i;  
  char betype;

  kbd.overlays_used = 0;

  if ((desc = XkbGetMap(display,XkbKeyBehaviorsMask,XkbUseCoreKbd)) == NULL) {
    TRACE_(layout)("XkbGetMap() failed to obtain keys behaviors, overlays unavailable!\n");
    return;
  }

  for (i = 1; i < MAX_OVERLAYS; i++)
    for (keyc = kbd.min_keycode; (keyc && (keyc <= kbd.max_keycode)); keyc++)
        kbd.keyc2over[i-1][keyc] = kbd.over2keyc[i-1][keyc] = keyc;
  
  for (keyc = kbd.min_keycode; (keyc && (keyc <= kbd.max_keycode)); keyc++) {
        betype = (desc->server->behaviors)[keyc].type;
        over   = (KeyCode)(desc->server->behaviors)[keyc].data;
        if (betype == XkbKB_Overlay1) {
           if (!(kbd.overlays_used & OVER1_ENABLED)) {
              TRACE_(layout)("Overlay1 enabled\n");
              kbd.overlays_used |= OVER1_ENABLED;
           }
           TRACE_(layout)("Keycode %d changed to %d (Overlay1)\n", keyc, over);
           if (kbd.over2keyc[OVER1_ENABLED-1][over] == over) {
             /* FIXME: case over<keyc ? */
             kbd.over2keyc[OVER1_ENABLED-1][over] = keyc;
             TRACE_(layout)("In Overlay1 keycode %d behaves as %d\n",
                       over, keyc);
           }
           if (kbd.over2keyc[OVER1_ENABLED+OVER2_ENABLED-1][over] == over) {
             /* FIXME: case over<keyc ? */
             kbd.over2keyc[OVER1_ENABLED+OVER2_ENABLED-1][over] = keyc;
             TRACE_(layout)("In Overlay3 keycode %d behaves as %d\n",
                       over, keyc);
           }
           kbd.keyc2over[OVER1_ENABLED-1][keyc] = 
           kbd.keyc2over[OVER1_ENABLED+OVER2_ENABLED-1][keyc] = over;
        } else
        if (betype == XkbKB_Overlay2) {
           if (!(kbd.overlays_used & OVER2_ENABLED)) {
              TRACE_(layout)("Overlay2 enabled\n");
              kbd.overlays_used |= OVER2_ENABLED;
           }
           TRACE_(layout)("Keycode %d changed to %d (Overlay2)\n", keyc, over);
           if (kbd.over2keyc[OVER2_ENABLED-1][over] == over) {
             /* FIXME: case over<keyc ? */
             kbd.over2keyc[OVER2_ENABLED-1][over] = keyc;
             TRACE_(layout)("In Overlay2 keycode %d behaves as %d\n",
                       over, keyc);
           }
           if (kbd.over2keyc[OVER1_ENABLED+OVER2_ENABLED-1][over] == over) {
             /* FIXME: case over<keyc ? */
             kbd.over2keyc[OVER1_ENABLED+OVER2_ENABLED-1][over] = keyc;
             TRACE_(layout)("In Overlay3 keycode %d behaves as %d\n",
                       over, keyc);
           }
           kbd.keyc2over[OVER2_ENABLED-1][keyc] = 
           kbd.keyc2over[OVER1_ENABLED+OVER2_ENABLED-1][keyc] = over;
        }
  }

  XkbFreeClientMap(desc, XkbAllMapComponentsMask, True);
  desc = NULL;
}
#endif /* HAVE_XKB */


/**********************************************************************
 *		KeycToOver + OverToKeyc
 *
 *    Helper functions that provide direct and inverse
 *   keycode transformation based on current overlay group  
 *
 */

inline static KeyCode
KeycToOver(KeyCode k){
#ifdef HAVE_XKB
    if      (use_xkb && kbd.overlay) return kbd.keyc2over[kbd.overlay-1][k];
    else
#endif /* HAVE_XKB */
         return k;
}

inline static KeyCode
OverToKeyc(KeyCode k){
#ifdef HAVE_XKB
    if      (use_xkb && kbd.overlay) return kbd.over2keyc[kbd.overlay-1][k];
    else
#endif /* HAVE_XKB */
         return k;
}

/**********************************************************************
 *		OverIgnore
 *
 *    Should we ignore this keycode when calculating the score
 *    for layout candidate? When overlay is not active, key that
 *    is mapped to by some overlay probably is virtual, so must
 *    not be taken into consideration. When overlay is active,
 *    we ignore keys that are mapped to other ones or mapped to
 *    by other overlay (thus it is better to have separate virtual
 *    keys for Overlay1 and Overlay2).
 */

inline static Bool
OverIgnore(KeyCode k){
#ifdef HAVE_XKB
    if  (use_xkb && kbd.overlays_used) {
      switch (kbd.overlay) {
        case 0: return ((kbd.over2keyc[0][k] != k)
                     || (kbd.over2keyc[1][k] != k));
        case 1: return ((kbd.keyc2over[0][k] != k)
                     || (kbd.over2keyc[1][k] != k));
        case 2: return ((kbd.keyc2over[1][k] != k)
                     || (kbd.over2keyc[0][k] != k));
        case 3: return ((kbd.keyc2over[1][k] != k)
                     || (kbd.keyc2over[0][k] != k));
        default: return 0; /* just to make compiler happy */
      }
    }
    else
#endif /* HAVE_XKB */
         return 0;
}


/**********************************************************************
 *		KEYBOARD_CharsForKeycode
 *
 *     Build arrays of keysyms and chars that key produces
 *     depending on shift level. If XKB is used, take it
 *     directly from keyboard description for the sake of
 *     speed and correctness.
 */

static Bool
KEYBOARD_CharsForKeyCode (Display *display,
                          int group,      /* ignored without XKB */
                          KeyCode keyc,        /*    input       */
                          KeySym (*ksym)[4],   /*    output      */
                          char (*chars)[4],    /* return values  */
                          char (*rawchars)[4], /* other codepage */
                          int *numchars, int *numrawchars) /* how many  */
{
  char Chars[5], RawChars[5];
  int i, w;
  KeySym keysym;
  int NumChars = 0, NumRawChars = 0;

#ifdef HAVE_XKB
  unsigned info;
  if (use_xkb) {
    w = XkbKeyNumGroups(desc,keyc);
    if (!w) {                                                 
      TRACE_(layout)("No groups for keycode %d\n", keyc);
      return False;
    }
    info = XkbOutOfRangeGroupInfo(XkbKeyGroupInfo(desc,keyc));
    if (group>=w)
      switch (info) {
        case (XkbRedirectIntoRange): group = XkbOutOfRangeGroupNumber(XkbKeyGroupInfo(desc,keyc));
                                   if (group >= w) group = 0; break;
        case (XkbClampIntoRange): ;  group = w-1; break;
        default: group %= w; break;
      }
    w = XkbKeyGroupWidth(desc,keyc,group);
  } else
#endif /* HAVE_XKB */
  w = (kbd.keysyms_per_keycode > 4) ? 4 : kbd.keysyms_per_keycode;

  for (i = 0; i < w; i++) {
#ifdef HAVE_XKB
    if (use_xkb) keysym=XkbKeySymEntry(desc, keyc, i, group);
    else
#endif
        keysym = XKeycodeToKeysym (display, keyc, i);
    if (ksym != NULL) (*ksym)[i] = keysym;
    if (!keysym) break; /* keysym == NoSymbol == 0 */
    if ((keysym < 0x8000) && (keysym != ' ')) {
          RawChars[i] = keysym & 0xFF;
#ifdef HAVE_XKB
          if (!use_xkb)
#endif
          Chars[i] = RawChars[i];
#ifdef HAVE_XKB
          else if (!XkbTranslateKeySym(display, &keysym, 0, &Chars[i], 1, NULL))
          {
                TRACE_(layout)("XKB could not translate keysym %ld\n", keysym);
                /* FIXME: query what keysym is used as Mode_switch, fill XKeyEvent
                 * with appropriate ShiftMask and Mode_switch, use XLookupString
                 * to get character in the local encoding.
                 */
                /* Chars[i] = keysym & 0xFF; */
                Chars[i] = 0;
          }
#endif
    }
    else Chars[i] = RawChars[i] = KEYBOARD_MapDeadKeysym(keysym);

    if (Chars[i] && (NumChars == i)) {
        if (chars) (*chars)[i]=Chars[i];
        NumChars++;
    }
    
    if (RawChars[i] && (NumRawChars == i)) {
        if (rawchars) (*rawchars)[i]=RawChars[i];
        NumRawChars++;
    }
    
    if ((NumChars != (i+1)) && (NumRawChars != (i+1))) break;
  }

  Chars[NumChars]='\0';
  RawChars[NumRawChars]='\0';
  if (numchars) (*numchars) = NumChars;
  if (numrawchars) (*numrawchars) = NumChars;
  if (chars && (NumChars < 4)) (*chars)[NumChars]='\0';
  if (rawchars && (NumRawChars < 4)) (*rawchars)[NumRawChars]='\0';

  TRACE_(layout)("Keycode %d translated to %d chars: \"%s\"\n", keyc, NumChars, Chars);
#ifdef HAVE_XKB
  if (use_xkb)
  TRACE_(layout)("              and to %d raw chars: \"%s\"\n", NumRawChars, RawChars);
#endif
  if (ksym)  TRACE_(layout)("  Keysyms: %s, %s, %s, %s\n",
            XKeysymToString((*ksym)[0]), XKeysymToString((*ksym)[1]),
            XKeysymToString((*ksym)[2]), XKeysymToString((*ksym)[3]));

  if (NumChars || NumRawChars) return True;
  else return False;
}

/**********************************************************************
 *		KEYBOARD_TestLayout
 *
 *     Get numerical description of how well the main_key_tab[current]
 *     layout matches the obtained keyc2chars 
 *
 *     NOTE: depends on kbd.overlay
 */

static void
KEYBOARD_TestLayout (const int current, const int is_raw, 
                     TempKeycData *keyc2char,
                     int *outscore, int *outmiss, int *outinv)
{
       int key, ok, match, prev, i;
       int score, miss, inv; /* Local copies */
       const char (*lkey)[MAIN_LEN][4];
       KeyCode keyc, over;
       score = match = miss = inv = prev = 0;
       lkey = main_key_tab[current].key;
        /* search for a match in layout table */
        /* right now, we just find an absolute match for defined positions */
        /* (undefined positions are ignored, so if it's defined as "3#" in */
        /* the table, it's okay that the X server has "3#", for example) */
        /* however, the score will be higher for longer matches */
       for (keyc = kbd.min_keycode; (keyc && (keyc <= kbd.max_keycode)); keyc++) {
         over = KeycToOver(keyc);
         if (!(*keyc2char)[is_raw][over][0] || OverIgnore(keyc)) continue;
         for (key = 0; key < MAIN_LEN; key++) {
         for (ok = i = 0; ((ok >= 0) && ((*lkey)[key][i])); i++) {
            if ((*lkey)[key][i] == (*keyc2char)[is_raw][over][i])  ok++;
             else ok = -1;
         }
          if (ok > 0) {
            score += ok;
            break;
          }
         }
        /* count the matches and mismatches */
         if (ok > 0) {
           match++;
/*           TRACE_(layout)("keycode %d matches entry %d\n", keyc, key); */
          /* and how much the keycode order matches */
           if (key < prev) inv++;
           prev = key;
         }
         else {
/*           TRACE_(layout)("mismatch for keycode %d\n", keyc); */
           miss++;
           for (i=0; i<4; i++) {
            if ((*keyc2char)[is_raw][over][i]) score--;
            else break;
           }
         }
      }
      TRACE_(layout)("\"%s\", %s lookup: matches=%d, mismatches=%d, inversions=%d, score=%d\n",
           main_key_tab[current].comment,
           (is_raw?"raw":"locale"),
           match, miss, inv, score);
      (*outscore) = score; (*outmiss) = miss; (*outinv) = inv;
}

/**********************************************************************
 *		KEYBOARD_DetectLayouts
 *
 *     Detect all layouts (1 without XKB, up to 16 if XKB is used)
 *     Make the list of preferred XKB and overlay groups for all
 *     layouts ( layout can be detected two or more times, so we should
 *     find the closest match for ActivateKeyboardLayout() )
 */

static void
KEYBOARD_DetectLayouts (void)
{
  Display *display = thread_display();
  int layout_no, current, saved, key, i, prev, have_chars, have_raw;
  int score, miss, inv, is_raw;
  int bestLayout, bestScore, bestInv, bestMiss, bestCP;
  WORD scan, vkey;

  const char (*lkey)[MAIN_LEN][4];
  TempKeycData *keyc2char[XkbNumKbdGroups];
  KeyCode keyc, over;
  KeySym keysym;
  XModifierKeymap *mmp;
  KeyCode *kcp;

#ifdef HAVE_XKB
  int save_group = 0, save_overlay = 0;
#endif /* HAVE_XKB */

  wine_tsx11_lock();
#ifdef HAVE_XKB
  if (use_xkb) {
    KEYBOARD_RefreshOverlays(display);
    /* get XKB keyboard description to speed up things later */
    if ((desc = XkbGetMap(display, XkbKeySymsMask | XkbKeyTypesMask, /* XkbAllMapComponentsMask, */
                          XkbUseCoreKbd))) {
        save_group   = kbd.group;
        save_overlay = kbd.overlay;
        kbd.num_layouts = MAX_OVERLAYS*kbd.num_groups;
    } else {
        ERR("cannot obtain XKB keyboard description from server, XKB use disabled!\n");
        kbd.num_groups = kbd.num_layouts = 1;
        use_xkb = kbd.overlays_used = 0;
    }
  } else
#endif /* HAVE_XKB */
  {
    TRACE("No XKB, keycodes from %d to %d\n",
          kbd.min_keycode, kbd.max_keycode);
    kbd.num_groups = kbd.num_layouts = 1;
#ifdef HAVE_XKB
    kbd.overlays_used = 0;
#endif
  }
  
  for (current=0; current<MAX_LAYOUTS; current++) {
     if (kbd.keyc[current]) free(kbd.keyc[current]);
     kbd.keyc[current] = NULL;
     kbd.layout[current] = -1;
     kbd.score[current]  = -1;
     kbd.cp[current] = 0;
  }

  /* Everything from the beginning */
    
  mmp = XGetModifierMapping(display);
  kcp = mmp->modifiermap;
  for (i = 0; i < 8; i += 1) /* There are 8 modifier keys */
  {
        int j;

        for (j = 0; j < mmp->max_keypermod; j += 1, kcp += 1)
            if (*kcp)
            {
                int k;
                for (k = 0; k < kbd.keysyms_per_keycode; k += 1)
                    if (XKeycodeToKeysym(display, *kcp, k) == XK_Num_Lock)
                    {
                        NumLockMask = 1 << i;
                        TRACE_(layout)("NumLockMask is %x\n", NumLockMask);
                    }
            }
  }
  XFreeModifiermap(mmp);

  /* Now store one keycode for each modifier. Used to simulate keypresses. */
  kcControl = XKeysymToKeycode(display, XK_Control_L);
  kcAlt = XKeysymToKeycode(display, XK_Alt_L);
  if (!kcAlt) kcAlt = XKeysymToKeycode(display, XK_Meta_L);
  kcShift = XKeysymToKeycode(display, XK_Shift_L);
  kcNumLock = XKeysymToKeycode(display, XK_Num_Lock);
  kcCapsLock = XKeysymToKeycode(display, XK_Caps_Lock);

  global_refresh = 0;  /* finished */

  /* Collect data about all possible layouts */
  for (current=0; current<kbd.num_groups; current++) {
      layout_no = current*MAX_OVERLAYS;
      if ((kbd.keyc[layout_no] = calloc(sizeof(KeycData)/sizeof(char),sizeof(char))) == NULL) {
          ERR("cannot allocate memory for permanent data for group #%d !\n",
             current);
          kbd.keyc[layout_no+1] =  kbd.keyc[layout_no+2] = kbd.keyc[layout_no+3] = NULL;
          kbd.layout[layout_no]   = kbd.layout[layout_no+1] = 
          kbd.layout[layout_no+1] = kbd.layout[layout_no+3] = -1;
          kbd.cp[layout_no]   = kbd.cp[layout_no+1] = 
          kbd.cp[layout_no+1] = kbd.cp[layout_no+3] = 0;
          keyc2char[current] = NULL;
          /* FIXME: Fallback */
          continue;
      }
      if ((keyc2char[current] = calloc(sizeof(TempKeycData)/sizeof(char),sizeof(char))) == NULL) {
          ERR("cannot allocate memory for temporary data for group #%d !\n",
             current);
          free(kbd.keyc[layout_no]);
          kbd.keyc[layout_no] = kbd.keyc[layout_no+1] =
          kbd.keyc[layout_no+2] = kbd.keyc[layout_no+3] = NULL;
          kbd.layout[layout_no]   = kbd.layout[layout_no+1] = 
          kbd.layout[layout_no+1] = kbd.layout[layout_no+3] = -1;
          kbd.cp[layout_no]   = kbd.cp[layout_no+1] = 
          kbd.cp[layout_no+1] = kbd.cp[layout_no+3] = 0;
          /* FIXME: Fallback */
          continue;
      }
      for (key=0; key<256; key++) {
         kbd.keyc[layout_no]->vkey[key] = 0; kbd.keyc[layout_no]->scan[key] = 0;
         for (i=0; i<4; i++) kbd.keyc[layout_no]->ksym[key][i] = 0;
      }
      TRACE("getting chars for layout #%d, group #%d :  keycodes from %d to %d\n",
              layout_no, current, kbd.min_keycode, kbd.max_keycode);
      for (keyc = kbd.min_keycode; (keyc && (keyc <= kbd.max_keycode)); keyc++) {
         if (!KEYBOARD_CharsForKeyCode (display, current,
                                    keyc, &kbd.keyc[layout_no]->ksym[keyc],
                                    &(*keyc2char[current])[0][keyc],
                                    &(*keyc2char[current])[1][keyc],
                                    &have_chars, &have_raw)) {
#ifdef HAVE_XKB
           if (use_xkb)
             TRACE_(layout)("No keysyms translated in group %d for keycode %d\n", current, keyc);
           else
#endif
             TRACE_(layout)("No keysyms translated for keycode %d\n", keyc);
         }
      }
#ifdef HAVE_XKB
      if (use_xkb) {
          for (i=1; i<MAX_OVERLAYS; i++) {
             layout_no = current*MAX_OVERLAYS+i;
             if ((i & kbd.overlays_used) != i) {
                TRACE("skip layout #%d (XKB group %d): overlay %d is incompatible with permitted %d\n",
                layout_no, current, i, kbd.overlays_used);
                kbd.keyc[layout_no] = NULL;
                kbd.layout[layout_no] = -1; kbd.cp[layout_no] = 0;
                continue;
             }
             if ((kbd.keyc[layout_no] = calloc(sizeof(KeycData)/sizeof(char),sizeof(char))) == NULL) {
                ERR("cannot allocate memory for permanent data for layout #%d, Overlay%d disabled!\n",
                    current,i);
                kbd.layout[layout_no]  = -1;
                kbd.cp[layout_no] = 0;
                kbd.overlays_used &= ~i;
                continue;
             }
             TRACE("copying keysyms for layout #%d : XKB group %d, overlay %d\n",
                layout_no, current, i);
             kbd.overlay = i;
             for (key=0; key<256; key++) {
                int level;
                keyc = KeycToOver(key);
                for (level=0; level<4; level++)
                   kbd.keyc[layout_no]->ksym[key][level] = kbd.keyc[current*MAX_OVERLAYS]->ksym[keyc][level];
                kbd.keyc[layout_no]->vkey[key] = 0; kbd.keyc[layout_no]->scan[key] = 0;
             }
          }
      }
  }
  if (use_xkb && desc) {
      XkbFreeClientMap(desc, XkbAllMapComponentsMask, True);
      desc = NULL;
#endif
  }
  wine_tsx11_unlock();

  /* Data collected, now detect layouts for all possible groups and overlays */
  for (layout_no=0; layout_no<kbd.num_layouts; layout_no++) {
      if (kbd.keyc[layout_no] == NULL) continue; /* Fallback action? */
#ifdef HAVE_XKB
      if (use_xkb) {
          kbd.group   = layout_no / MAX_OVERLAYS;
          kbd.overlay = layout_no % MAX_OVERLAYS;
          TRACE("examining layout #%d : XKB group %d, overlay %d, keycodes from %d to %d\n",
             layout_no, kbd.group, kbd.overlay, kbd.min_keycode, kbd.max_keycode);
      }
#endif

  bestLayout = bestScore = bestInv = bestMiss = bestCP = 0;

  /* First go through all already detected layouts */
  for (saved = 0; saved < MAX_REMEMBER; saved++) {
       TRACE_(layout)("Compare with a previously detected #%d : %d / %d / %d \n",
              saved,
              kbd.prev[saved][0].layout,
              kbd.prev[saved][0].score,
              kbd.prev[saved][0].inv);
       if (kbd.prev[saved][0].layout == -1) break;
       KEYBOARD_TestLayout(kbd.prev[saved][0].layout,
                           kbd.prev[saved][0].is_raw,
                           keyc2char[kbd.group], &score, &miss, &inv);
       if ((score != kbd.prev[saved][0].score) ||
           (inv != kbd.prev[saved][0].inv)) continue;
       for (i = 1; i<MAX_NEIGHBORS; i++) {
           if (kbd.prev[saved][i].layout != -1) {
               i = MAX_NEIGHBORS; break;
           }
/*           TRACE_(layout)("Compare with an alternative #%d/%d  : %d / %d / %d \n",
                      saved, i,
                      kbd.prev[saved][i].layout,
                      kbd.prev[saved][i].score,
                      kbd.prev[saved][i].inv);
*/
           KEYBOARD_TestLayout(kbd.prev[saved][i].layout,
                           kbd.prev[saved][i].is_raw,
                           keyc2char[kbd.group], &score, NULL, &inv);
           if ((score != kbd.prev[saved][i].score) ||
               (inv != kbd.prev[saved][i].inv)) break;
       }
       if (i == MAX_NEIGHBORS) { /* Found; "saved" keeps the matching layout */
           TRACE("Layout #%d seems to be a previously detected \"%s\", %s lookup\n",
                      layout_no,
                      main_key_tab[kbd.prev[saved][0].layout].comment,
                     (kbd.prev[saved][0].is_raw?"raw":"locale"));
           bestLayout = kbd.prev[saved][0].layout;
           bestScore  = kbd.prev[saved][0].score;
           bestInv    = kbd.prev[saved][0].inv;
           bestMiss   = miss;
           bestCP     = kbd.prev[saved][0].is_raw;
           goto found;
       }
  }
  /* None of previous layouts matched */

  if (saved == MAX_REMEMBER) { /* Delete the oldest detected layout*/
       for (saved = 0; saved<MAX_REMEMBER-1; saved++) {
           for (i = 0; i<MAX_NEIGHBORS; i++) {
                kbd.prev[saved][i].layout = kbd.prev[saved+1][i].layout;
                kbd.prev[saved][i].score  = kbd.prev[saved+1][i].score;
                kbd.prev[saved][i].inv    = kbd.prev[saved+1][i].inv;
                kbd.prev[saved][i].is_raw = kbd.prev[saved+1][i].is_raw;
           }
       }
       for (i = 0; i<MAX_NEIGHBORS; i++) {
                kbd.prev[saved][i].layout = 0;
                kbd.prev[saved][i].score  = 0;
                kbd.prev[saved][i].inv    = 0;
                kbd.prev[saved][i].is_raw = 0;
       }
       saved = MAX_REMEMBER-1;
  }  

  /*
     Match against all layouts known to Wine,
     write MAX_NEIGHBORS best candidates into
     a row kbd.prev[saved][...]
   */
  for (current = 0; main_key_tab[current].comment; current++) {
      is_raw = 0; /* first try locale encoding */
    again:
      KEYBOARD_TestLayout(current, is_raw, keyc2char[kbd.group], &score, &miss, &inv);
      for (i = 0; i<MAX_NEIGHBORS; i++) {
          if ((!kbd.prev[saved][i].layout) || /* no layout */
              (score>kbd.prev[saved][i].score) ||  /* score is better */
             ((score==kbd.prev[saved][i].score) && 
              (inv<kbd.prev[saved][i].inv)) ||     /* less swapped pairs */
             ((score==kbd.prev[saved][i].score) && 
              (inv==kbd.prev[saved][i].inv) &&     /* locale lookup is preferred */
              (!is_raw) && kbd.prev[saved][i].is_raw)) {
             /* Here we stop and insert a candidate in the list */
            for (prev = MAX_NEIGHBORS-1; prev>i; prev--) {
                    kbd.prev[saved][prev].layout = kbd.prev[saved][prev-1].layout;
                    kbd.prev[saved][prev].score  = kbd.prev[saved][prev-1].score;
                    kbd.prev[saved][prev].inv    = kbd.prev[saved][prev-1].inv;
                    kbd.prev[saved][prev].is_raw = kbd.prev[saved][prev-1].is_raw;
            }
            kbd.prev[saved][i].layout = current;
            kbd.prev[saved][i].score  = score;
            kbd.prev[saved][i].inv    = inv;
            kbd.prev[saved][i].is_raw = is_raw;
            if (!i) {
                bestLayout = current;
                bestScore  = score;
                bestInv    = inv;
                bestMiss   = miss;
                bestCP     = is_raw;
            }
            break;
         } 
      }
      if (!is_raw) {
          is_raw = 1;
          goto again; /* now simply get lower bytes */
      }
  } /* End of detection of a single layout */
    
  found:
  if (bestCP) 
         bestCP = main_key_tab[bestLayout].codepage;
  if (bestMiss)
    TRACE("Detected layout #%d is \"%s\" (closest match), codepage %d\n",
                     layout_no,
                     main_key_tab[bestLayout].comment,
                     bestCP);
  else
    TRACE("Detected layout #%d is \"%s\" (exact match), codepage %d\n",
                     layout_no,
                     main_key_tab[bestLayout].comment,
                     bestCP);

  /* Is this layout forced in the registry? */
  if (reg_layout[layout_no] != -1) {
        is_raw = (reg_cp[layout_no] > 0);
        KEYBOARD_TestLayout(reg_layout[layout_no], is_raw,
           keyc2char[kbd.group], &score, &miss, &inv);
        TRACE("Overridden by registry settings, layout is \"%s\", codepage %d,\n",
                     main_key_tab[reg_layout[layout_no]].comment,
                     reg_cp[layout_no]);
        TRACE("                                 score %d (best is %d)\n",
                     score, bestScore);
        bestScore = score;
        kbd.layout[layout_no] = reg_layout[layout_no];
        kbd.cp[layout_no]     = reg_cp[layout_no];
  } else { /* Use the best found candidate */
    kbd.layout[layout_no] = bestLayout;
    if (reg_cp[layout_no] != -1) {
           TRACE("Codepage changed by registry settings to %d\n",
           reg_cp[layout_no]);
           kbd.cp[layout_no] = reg_cp[layout_no];
    } else
           kbd.cp[layout_no] = bestCP;
  }

#ifdef HAVE_XKB
  if (use_xkb && (bestScore>0)) {
      kbd.pref[layout_no] = bestScore;
      for (i=0; i<layout_no; i++) {
        key = i % MAX_OVERLAYS;
        if ((key & kbd.overlays_used) != key) continue;
        if (kbd.pref[i] && (main_key_tab[kbd.layout[i]].lcid ==
            main_key_tab[kbd.layout[layout_no]].lcid)) {
          if (kbd.pref[layout_no] > kbd.pref[i])
            kbd.pref[i] = 0;
          else
            kbd.pref[layout_no] = 0;
          break;
        }
      }
  } else
#endif /* HAVE_XKB */
  kbd.pref[layout_no] = 0;

  for (keyc = kbd.min_keycode; (keyc && (keyc <= kbd.max_keycode)); keyc++)
  {
        over = KeycToOver(keyc);
        keysym = kbd.keyc[layout_no]->ksym[over][0];
        is_raw = (kbd.cp[layout_no] != 0);
        have_chars = (*keyc2char[kbd.group])[is_raw][over][0];
        vkey = 0; scan = 0;
        if (keysym)  /* otherwise, keycode not used */
        {
            if ((keysym >> 8) == 0xFF)         /* non-character key */
            {
                vkey = nonchar_key_vkey[keysym & 0xff];
                scan = nonchar_key_scan[keysym & 0xff];
                /* set extended bit when necessary */
                if (scan & 0x100) vkey |= 0x100;
            } else if ((keysym >> 8) == 0x1008FF) { /* XFree86 vendor keys */
                vkey = xfree86_vendor_key_vkey[keysym & 0xff];
                /* All vendor keys are extended with a scan code of 0 per testing on WinXP */
                scan = 0x100;
                vkey |= 0x100;
            } else if (keysym == 0x20) {                 /* Spacebar */
                vkey = VK_SPACE;
                scan = 0x39;
            } else if (have_chars) {
              /* we seem to need to search the layout-dependent scancodes */
              lkey = main_key_tab[kbd.layout[layout_no]].key;
              int maxlen=0,maxval=-1, ok;
              /* find key with longest match streak */
              for (key=0; key<MAIN_LEN; key++) {
                for(i=0, ok=1; i<4; i++) {
                  if (!(*lkey)[key][i]) break;
                  if ((*lkey)[key][i]!=(*keyc2char[kbd.group])[is_raw][over][i]) {
                   ok=0; break;
                  }
                }
                if ((ok && i)||(i>maxlen)) {
                  maxlen=i; maxval=key;
                }
                if (ok) break;
              }
              if (maxval>=0) {
                /* got it */
                const WORD (*lscan)[MAIN_LEN] = main_key_tab[kbd.layout[layout_no]].scan;
                const WORD (*lvkey)[MAIN_LEN] = main_key_tab[kbd.layout[layout_no]].vkey;
                scan = (*lscan)[maxval];
                vkey = (*lvkey)[maxval];
                unsigned char bkey[5];
                bkey[0]=(*keyc2char[kbd.group])[is_raw][over][0];
                bkey[1]=(*keyc2char[kbd.group])[is_raw][over][1];
                bkey[2]=(*keyc2char[kbd.group])[is_raw][over][2];
                bkey[3]=(*keyc2char[kbd.group])[is_raw][over][3];
                bkey[4] = '\0';
                TRACE_(layout)("Identified %d chars of \"%s\" / "
                            "\"%02x,%02x,%02x,%02x\" as \"%s\", entry %d\n",
                            maxlen, bkey, bkey[0], bkey[1], bkey[2], bkey[3],
                            (*lkey)[maxval], maxval); 
              }
            }

        }
        TRACE_(layout)("keycode %04x => vkey %04x\n", keyc, vkey);
        kbd.keyc[layout_no]->vkey[keyc] = vkey;
        kbd.keyc[layout_no]->scan[keyc] = scan;
  } /* for */

    /* If some keys still lack scancodes, assign some arbitrary ones to them now */
  for (scan = 0x60, keyc = kbd.min_keycode; (keyc && (keyc <= kbd.max_keycode)); keyc++)
      if (kbd.keyc[layout_no]->vkey[keyc] && !kbd.keyc[layout_no]->scan[keyc]) {
        const char *ksname;
        wine_tsx11_lock();
        keysym = XKeycodeToKeysym(display, keyc, 0);
        ksname = XKeysymToString(keysym);
        if (!ksname) ksname = "NoSymbol";

        /* should make sure the scancode is unassigned here, but >=0x60 currently always is */

        TRACE_(layout)("assigning scancode %02x to unidentified keycode %02x (%s)\n",scan,keyc,ksname);
        kbd.keyc[layout_no]->scan[keyc]=scan++;
        wine_tsx11_unlock();
      }
  } 
  /* End of grand cycle, free temp. memory, restore current group and overlay */
  for (kbd.group=0; kbd.group<kbd.num_groups; kbd.group++) {
     if (keyc2char[kbd.group]) free(keyc2char[kbd.group]);
  }
#ifdef HAVE_XKB
  if (use_xkb) {
    for (layout_no=0; layout_no<kbd.num_layouts; layout_no++) {
         if (kbd.pref[layout_no]) {
           TRACE("Preferred layout #%d \"%s\", score %d, codepage %d\n", layout_no,
                main_key_tab[kbd.layout[layout_no]].comment,
                kbd.pref[layout_no],
                kbd.cp[layout_no]);
         }
    }
    kbd.group   = save_group;
    kbd.overlay = save_overlay & kbd.overlays_used;
    kbd.layout_no = MAX_OVERLAYS*kbd.group + kbd.overlay;
  } else
#endif
  kbd.layout_no = 0;
}

/**********************************************************************
 *		InitKeyboard (X11DRV.@)
 */
void X11DRV_InitKeyboard(void)
{
    int current, no;
    const int buf_size = 128;
    unsigned int cp; 
    DWORD type, count;
    char buffer[buf_size], name_buffer[12];
    HKEY hkey;


    kbd.min_keycode = 0;
    kbd.max_keycode = 255;
    kbd.keysyms_per_keycode = 4;
    for (no=0; no<MAX_LAYOUTS; no++) {
        kbd.layout[no] = -1;
        kbd.score[no] = kbd.cp[no] = 0;
        kbd.keyc[no] = NULL;
    }
    for (no=0; no<MAX_REMEMBER; no++) {
        for (current=0; current<MAX_NEIGHBORS; current++) {
            kbd.prev[no][current].layout = -1;
            kbd.prev[no][current].score = kbd.prev[no][current].inv = \
            kbd.prev[no][current].is_raw = 0;
        }
    }

    /* @@ Wine registry key: HKCU\Software\Wine\X11 Driver\Keyboard */
    if (RegOpenKeyA(HKEY_CURRENT_USER, INIKeyboard, &hkey)) hkey = 0;

    for (no=0; no<(use_xkb ? MAX_LAYOUTS : 1); no++ ) {
        reg_layout[no] = -1; reg_cp[no] = -1;
        cp = 0;
        if (!hkey) continue;

        sprintf(name_buffer, "Layout%d", no);
        count = buf_size;
        buffer[0] = 0;
        RegQueryValueExA(hkey, name_buffer, 0, &type, (LPBYTE)buffer, &count);
        if (!buffer[0]) continue;
        for (current=0; main_key_tab[current].comment; current++)
            if (!strcmp(buffer,main_key_tab[current].comment)) break;
        if (!main_key_tab[current].comment) {
            TRACE("Layout #%d \"%s\" specified in registry but not found\n",
                no, buffer);
            continue;
        }
        TRACE("Layout #%d \"%s\" specified in registry\n", no, buffer);
        reg_layout[no] = current;

        count = buf_size;
        buffer[0] = 0;
        sprintf(name_buffer, "Codepage%d", no);
        RegQueryValueExA(hkey, name_buffer, 0, &type, (LPBYTE)buffer, &count);
        if (buffer[0]) cp=atoi(buffer);
        /* FIXME: is this codepage valid? */
        if (cp>=0) {
          reg_cp[no] = cp;
          TRACE("Codepage %d for layout #%d specified in registry\n", cp, no);
        }
    }
    if (hkey) RegCloseKey(hkey);

    for (current = 0; main_key_tab[current].comment; current++) {
       if (!use_xkb) continue;
       if (main_key_tab[current].codepage)
                TRACE_(layout)("keysyms codepage for %s defaults to %d\n",
                    main_key_tab[current].comment,main_key_tab[current].codepage);
       else {
              if (!GetLocaleInfoW(main_key_tab[current].lcid,
                LOCALE_IDEFAULTUNIXCODEPAGE|LOCALE_RETURN_NUMBER,
                (WCHAR *)(&cp),sizeof(cp)/sizeof(WCHAR))) {
                TRACE_(layout)("keysyms codepage for %s unknown, cannot get DefaultUnixCodepage, skipping\n",
                    main_key_tab[current].comment);
              } else {
                TRACE_(layout)("set keysyms codepage for %s to DefaultUnixCodepage = %d\n",
                    main_key_tab[current].comment,cp);
                main_key_tab[current].codepage = cp;
              }
       }
    }

    wine_tsx11_lock();
    KEYBOARD_LayoutSwitched(); /* get current group and overlay */
    wine_tsx11_unlock();

    KEYBOARD_DetectLayouts(); /* detect all layouts */
}


/**********************************************************************
 *		GetAsyncKeyState (X11DRV.@)
 */
SHORT X11DRV_GetAsyncKeyState(INT key)
{
    SHORT retval;

    /* Photoshop livelocks unless mouse events are included here */
    X11DRV_MsgWaitForMultipleObjectsEx( 0, NULL, 0, QS_KEY | QS_MOUSE, 0 );

    retval = ((key_state_table[key] & 0x40) ? 0x0001 : 0) |
             ((key_state_table[key] & 0x80) ? 0x8000 : 0);
    key_state_table[key] &= ~0x40;
    TRACE_(key)("(%x) -> %x\n", key, retval);
    return retval;
}


/***********************************************************************
 *		GetKeyboardLayoutList (X11DRV.@)
 */
UINT X11DRV_GetKeyboardLayoutList(INT size, HKL *hkl)
{
    INT i;

    TRACE("%d, %p\n", size, hkl);

    if (!size)
    {
        size = 4096; /* hope we will never have that many */
        hkl = NULL;
    }

    for (i = 0; main_key_tab[i].comment && (i < size); i++)
    {
        if (hkl)
        {
            ULONG_PTR layout = main_key_tab[i].lcid;
            LANGID langid;

            /* see comment for GetKeyboardLayout */
            langid = PRIMARYLANGID(LANGIDFROMLCID(layout));
            if (langid == LANG_CHINESE || langid == LANG_JAPANESE || langid == LANG_KOREAN)
                layout |= 0xe001 << 16; /* FIXME */
            else
                layout |= layout << 16;

            hkl[i] = (HKL)layout;
        }
    }
    return i;
}


/***********************************************************************
 *		GetKeyboardLayout (X11DRV.@)
 */
HKL X11DRV_GetKeyboardLayout(DWORD dwThreadid)
{
    ULONG_PTR layout;
    LANGID langid;

    if (dwThreadid && dwThreadid != GetCurrentThreadId())
        FIXME("couldn't return keyboard layout for thread %04x\n", dwThreadid);

#if 1
    layout = main_key_tab[kbd.layout[kbd.layout_no]].lcid;
#else
    /* FIXME:
     * Winword uses return value of GetKeyboardLayout as a codepage
     * to translate ANSI keyboard messages to unicode. But we have
     * a problem with it: for instance Polish keyboard layout is
     * identical to the US one, and therefore instead of the Polish
     * locale id we return the US one.
     */
    layout = GetUserDefaultLCID();
#endif
    /* 
     * Microsoft Office expects this value to be something specific
     * for Japanese and Korean Windows with an IME the value is 0xe001
     * We should probably check to see if an IME exists and if so then
     * set this word properly.
     */
    langid = PRIMARYLANGID(LANGIDFROMLCID(layout));
    if (langid == LANG_CHINESE || langid == LANG_JAPANESE || langid == LANG_KOREAN)
        layout |= 0xe001 << 16; /* FIXME */
    else
        layout |= layout << 16;

    return (HKL)layout;
}


/***********************************************************************
 *		GetKeyboardLayoutName (X11DRV.@)
 */
BOOL X11DRV_GetKeyboardLayoutName(LPWSTR name)
{
    static const WCHAR formatW[] = {'%','0','8','l','x',0};
    DWORD layout;
    LANGID langid;

    layout = main_key_tab[kbd.layout[kbd.layout_no]].lcid;

    /* see comment for GetKeyboardLayout */
    langid = PRIMARYLANGID(LANGIDFROMLCID(layout));
    if (langid == LANG_CHINESE || langid == LANG_JAPANESE || langid == LANG_KOREAN)
        layout |= 0xe001 << 16; /* FIXME */
    else
        layout |= layout << 16;

    sprintfW(name, formatW, layout);
    TRACE("returning %s\n", debugstr_w(name));
    return TRUE;
}


/***********************************************************************
 *		LoadKeyboardLayout (X11DRV.@)
 */
HKL X11DRV_LoadKeyboardLayout(LPCWSTR name, UINT flags)
{
    FIXME("%s, %04x: stub!\n", debugstr_w(name), flags);
    SetLastError(ERROR_CALL_NOT_IMPLEMENTED);
    return 0;
}


/***********************************************************************
 *		UnloadKeyboardLayout (X11DRV.@)
 */
BOOL X11DRV_UnloadKeyboardLayout(HKL hkl)
{
    FIXME("%p: stub!\n", hkl);
    SetLastError(ERROR_CALL_NOT_IMPLEMENTED);
    return FALSE;
}


/***********************************************************************
 *		ActivateKeyboardLayout (X11DRV.@)
 */
HKL X11DRV_ActivateKeyboardLayout(HKL hkl, UINT flags)
{
#ifdef HAVE_XKB
      /* Ignore all flags : */
      /* KLF_REORDER, KLF_RESET, KLF_SETFORPROCESS, KLF_SHIFTLOCK */
      /* KLF_UNLOADPREVIOUS - unsupported by Windows */
    if (use_xkb) {
      HKL cur_lcid;
      ULONG_PTR lcid;
      int i;
      Display *display;
      wine_tsx11_lock();
      display = thread_display();
      lcid = (ULONG_PTR)hkl;
      cur_lcid = (HKL)(main_key_tab[kbd.layout[kbd.layout_no]].lcid);
      if (lcid == HKL_NEXT) {
        for(i=kbd.layout_no+1; i != kbd.layout_no; i++) {
          if (i == kbd.num_layouts) i=0;
          if (kbd.pref[i]) break;
        }
        if (i == kbd.layout_no) {
          TRACE("Failed to activate next layout\n");
          wine_tsx11_unlock();
          return 0;
        }
      } else
      if (lcid == HKL_PREV) {
        for(i=kbd.layout_no-1; i != kbd.layout_no; i--) {
          if (i < 0) i=kbd.num_layouts-1;
          if (kbd.pref[i]) break;
        }
        if (i == kbd.layout_no) {
          TRACE("Failed to activate previous layout\n");
          wine_tsx11_unlock();
          return 0;
        }
      } else {
        if (main_key_tab[kbd.layout[kbd.layout_no]].lcid == lcid) {
          TRACE("Handle %08lx is already active\n",lcid);
          wine_tsx11_unlock();
          return hkl;
        }
        for(i=0; i<kbd.num_layouts; i++)
          if (kbd.pref[i] && 
           (main_key_tab[kbd.layout[i]].lcid == lcid)) break;
        if (i == kbd.num_layouts) {
          TRACE("Failed to activate handle %08lx\n",lcid);
          wine_tsx11_unlock();
          return 0;
        }
      }
      XkbLockGroup(display, XkbUseCoreKbd, i/MAX_OVERLAYS);
      XkbChangeEnabledControls(display, XkbUseCoreKbd, 
           XkbOverlay1Mask | XkbOverlay2Mask ,
           (((i%MAX_OVERLAYS) & OVER1_ENABLED ?  XkbOverlay1Mask : 0) |
            ((i%MAX_OVERLAYS) & OVER2_ENABLED ?  XkbOverlay2Mask : 0)));
/*    KEYBOARD_LayoutSwitched(); // Should we wait for server response ? */
      TRACE(" #%d, \"%s\", handle %08lx\n",
         kbd.layout_no, main_key_tab[kbd.layout[kbd.layout_no]].comment, lcid);
      wine_tsx11_unlock();
      return cur_lcid;
    } else
#endif /* HAVE_XKB */
    {
      FIXME("%p, %04x: stub: not implemented without XKB!\n", hkl, flags);
      SetLastError(ERROR_CALL_NOT_IMPLEMENTED);
      return 0;
    }
}


/***********************************************************************
 *           X11DRV_MappingNotify
 */
void X11DRV_MappingNotify( HWND dummy, XEvent *event )
{
    HWND hwnd;

    wine_tsx11_lock();
    XRefreshKeyboardMapping(&event->xmapping);
    wine_tsx11_unlock();
    X11DRV_InitKeyboard();

    hwnd = GetFocus();
    if (!hwnd) hwnd = GetActiveWindow();
    PostMessageW(hwnd, WM_INPUTLANGCHANGEREQUEST,
                 0 /*FIXME*/, (LPARAM)X11DRV_GetKeyboardLayout(0));
}


/***********************************************************************
 *		VkKeyScanEx (X11DRV.@)
 *
 * Note: Windows ignores HKL parameter and uses current active layout instead
 */
SHORT X11DRV_VkKeyScanEx(WCHAR wChar, HKL hkl)
{
    Display *display = thread_display();
    KeyCode keycode;
    KeySym keysym;
    int i, index;
    CHAR cChar;
    SHORT ret;
    int layout_no = kbd.layout_no;

#if 0 /* if the note above is correct ... */
    if (hkl != (HKL)(main_key_tab[kbd.layout[layout_no]].lcid)) {
#ifdef HAVE_XKB 
        for(layout_no=0; layout_no<kbd.num_layouts; layout_no++) {
           if (!kbd.pref[layout_no]) continue;
           if (hkl == (HKL)(main_key_tab[kbd.layout[layout_no]].lcid)) break;
        }
    }
    if (layout_no == kbd.num_layouts) {
        layout_no = kbd.layout_no;
#endif /* HAVE_XKB */
        TRACE_(key)("keyboard layout %p is not supported\n", hkl);
    }
#endif /* 0 */

    /* FIXME: what happens if wChar is not a Latin1 character and CP_UNIXCP
     * is UTF-8 (multibyte encoding)?
     */
    if (kbd.cp[layout_no]) {
        if (!WideCharToMultiByte(kbd.cp[layout_no], 0, &wChar, 1, &cChar, 1, NULL, NULL))
        {
            WARN("no translation from unicode to codepage %d for 0x%02x\n", kbd.cp[layout_no], wChar);
            return -1;
        }
    } else 
    {
        if (!WideCharToMultiByte(CP_UNIXCP, 0, &wChar, 1, &cChar, 1, NULL, NULL))
        {
            WARN("no translation from unicode to CP_UNIXCP for 0x%02x\n", wChar);
            return -1;
        }
    }

    TRACE_(key)("wChar 0x%02x -> cChar '%c'\n", wChar, cChar);

    /* char->keysym (same for ANSI chars) */
    keysym = (unsigned char)cChar; /* (!) cChar is signed */
    if (keysym <= 27) keysym += 0xFF00; /* special chars : return, backspace... */

    wine_tsx11_lock();
    keycode = XKeysymToKeycode(display, keysym);  /* keysym -> keycode */
    if (!keycode)
    {
        if (keysym >= 0xFF00) /* Windows returns 0x0240 + cChar in this case */
        {
            ret = 0x0240 + cChar; /* 0x0200 indicates a control character */
            TRACE_(key)(" ... returning ctrl char %#.2x\n", ret);
            wine_tsx11_unlock();
            return ret;
        }
        /* It didn't work ... let's try with deadchar code. */
        TRACE_(key)("retrying with | 0xFE00\n");
        keycode = XKeysymToKeycode(display, keysym | 0xFE00);
    }
    wine_tsx11_unlock();

    TRACE_(key)("'%c'(%#lx, %lu): got keycode %#.2x (%d)\n",
            cChar, keysym, keysym, keycode, keycode);

    /* keycode -> (keyc2vkey) vkey */
    ret = kbd.keyc[layout_no]->vkey[keycode];

    if (!keycode || !ret)
    {
        TRACE_(key)("keycode for '%c' not found, returning -1\n", cChar);
        return -1;
    }

    index = -1;
    for (i = 0; i < 4; i++) /* find shift state */
    {
      if (kbd.keyc[layout_no]->ksym[keycode][i] && 
             (kbd.keyc[layout_no]->ksym[keycode][i] == keysym)) {
         index = i;
         break;
      }
    }

    switch (index)
    {
        default:
        case -1:
            WARN_(key)("Keysym %lx not found while parsing the keycode table\n", keysym);
            return -1;

        case 0: break;
        case 1: ret += 0x0100; break;
        case 2: ret += 0x0600; break;
        case 3: ret += 0x0700; break;
    }
    /*
      index : 0     adds 0x0000
      index : 1     adds 0x0100 (shift)
      index : ?     adds 0x0200 (ctrl)
      index : 2     adds 0x0600 (ctrl+alt)
      index : 3     adds 0x0700 (ctrl+alt+shift)
     */

    TRACE_(key)(" ... returning %#.2x\n", ret);
    return ret;
}

/***********************************************************************
 *		MapVirtualKeyEx (X11DRV.@)
 */
UINT X11DRV_MapVirtualKeyEx(UINT wCode, UINT wMapType, HKL hkl)
{
    int layout_no = kbd.layout_no;
    Display *display = thread_display();

#define returnMVK(value) { TRACE("returning 0x%x.\n",value); return value; }

    TRACE_(key)("wCode=0x%x, wMapType=%d, hkl %p\n", wCode, wMapType, hkl);
    if (hkl != (HKL)(main_key_tab[kbd.layout[layout_no]].lcid)) {
#ifdef HAVE_XKB 
        for(layout_no=0; layout_no<kbd.num_layouts; layout_no++) {
           if (!kbd.pref[layout_no]) continue;
           if (hkl == (HKL)(main_key_tab[kbd.layout[layout_no]].lcid)) break;
        }
    }
    if (layout_no == kbd.num_layouts) {
        layout_no = kbd.layout_no;
#endif /* HAVE_XKB */
        TRACE_(key)("keyboard layout %p is not supported\n", hkl);
    }

        switch(wMapType) {
                case 0: { /* vkey-code to scan-code */
                        /* let's do vkey -> keycode -> scan */
                        int keyc;
                        for (keyc=kbd.min_keycode; keyc<=kbd.max_keycode; keyc++)
                                if ((kbd.keyc[layout_no]->vkey[keyc] & 0xFF) == wCode)
                                        returnMVK (kbd.keyc[layout_no]->scan[keyc] & 0xFF);
                        TRACE_(key)("returning no scan-code.\n");
                        return 0; }

                case 1: { /* scan-code to vkey-code */
                        /* let's do scan -> keycode -> vkey */
                        int keyc;
                        for (keyc=kbd.min_keycode; keyc<=kbd.max_keycode; keyc++)
                                if ((kbd.keyc[layout_no]->scan[keyc] & 0xFF) == (wCode & 0xFF))
                                        returnMVK (kbd.keyc[layout_no]->vkey[keyc] & 0xFF);
                        TRACE_(key)("returning no vkey-code.\n");
                        return 0; }

                case 2: { /* vkey-code to unshifted ANSI code */
                        /* we still don't know what "unshifted" means. in windows VK_W (0x57)
                         * returns 0x57, which is upercase 'W'. So we have to return the uppercase
                         * key.. Looks like something is wrong with the MS docs?
                         * This is only true for letters, for example VK_0 returns '0' not ')'.
                         * - hence we use the lock mask to ensure this happens.
                         */
                        /* let's do vkey -> keycode -> (XLookupString) ansi char */
                        XKeyEvent e;
                        KeySym keysym = 0;
                        int keyc;
                        char s[2];
                        e.display = display;

                        e.state = LockMask;
                        /* LockMask should behave exactly like caps lock - upercase
                         * the letter keys and thats about it. */

                        wine_tsx11_lock();

                        e.keycode = 0;
                        /* We exit on the first keycode found, to speed up the thing. */
                        for (keyc=kbd.min_keycode; (keyc<=kbd.max_keycode) && (!e.keycode) ; keyc++)
                        { /* Find a keycode that could have generated this virtual key */
                            if  ((kbd.keyc[layout_no]->vkey[keyc] & 0xFF) == wCode)
                            { /* We filter the extended bit, we don't know it */
                                e.keycode = KeycToOver(keyc); /* Store it temporarily */
                                if ( keyc != e.keycode )
                                   TRACE_(key)("Keycode %02x turned back to %02x\n", keyc, e.keycode);
                                if ((EVENT_event_to_vkey(0,&e) & 0xFF) != wCode) {
                                    e.keycode = 0; /* Wrong one (ex: because of the NumLock
                                         state), so set it to 0, we'll find another one */
                                }
                                else {
                                    e.keycode = KeycToOver(keyc); /* again */
                                }
                            }
                        }

                        if ((wCode>=VK_NUMPAD0) && (wCode<=VK_NUMPAD9))
                          e.keycode = XKeysymToKeycode(e.display, wCode-VK_NUMPAD0+XK_KP_0);

                        if (wCode==VK_DECIMAL)
                          e.keycode = XKeysymToKeycode(e.display, XK_KP_Decimal);

                        if (!e.keycode)
                        {
                          WARN_(key)("Unknown virtual key %X !!!\n", wCode);
                          wine_tsx11_unlock();
                          return 0; /* whatever */
                        }
                        TRACE_(key)("Found keycode %d (0x%2X)\n",e.keycode,e.keycode);

                        if (XLookupString(&e, s, 2, &keysym, NULL) && !kbd.cp[layout_no])
                        {
                            wine_tsx11_unlock();
                            returnMVK (*s);
                        } 
                        else if (keysym && kbd.cp[layout_no]) {
                            KeySym lkeysym, ukeysym;
                            XConvertCase(keysym,&lkeysym,&ukeysym);
                            wine_tsx11_unlock();
                            if (ukeysym < 0x8000)
                                returnMVK ((unsigned)(ukeysym & 0xFF)); /* ???? */
                        }
                        TRACE_(key)("returning no ANSI.\n");
                        wine_tsx11_unlock();
                        return 0;
                        }

                case 3:   /* **NT only** scan-code to vkey-code but distinguish between  */
                          /*             left and right  */
                          FIXME(" stub for NT\n");
                          return 0;

                default: /* reserved */
                        WARN_(key)("Unknown wMapType %d !\n", wMapType);
                        return 0;
        }
        return 0;
}

/***********************************************************************
 *		GetKeyNameText (X11DRV.@)
 */
INT X11DRV_GetKeyNameText(LONG lParam, LPWSTR lpBuffer, INT nSize)
{
  int vkey, ansi, scanCode;
  KeyCode keyc;
  int keyi;
  KeySym keys;
  char *name;

  scanCode = lParam >> 16;
  scanCode &= 0x1ff;  /* keep "extended-key" flag with code */

  /* FIXME: should use MVK type 3 (NT version that distinguishes right and left */
  vkey = X11DRV_MapVirtualKeyEx(scanCode, 1, X11DRV_GetKeyboardLayout(0));

  /*  handle "don't care" bit (0x02000000) */
  if (!(lParam & 0x02000000)) {
    switch (vkey) {
         case VK_LSHIFT:
         case VK_RSHIFT:
                          vkey = VK_SHIFT;
                          break;
       case VK_LCONTROL:
       case VK_RCONTROL:
                          vkey = VK_CONTROL;
                          break;
          case VK_LMENU:
          case VK_RMENU:
                          vkey = VK_MENU;
                          break;
               default:
                          break;
    }
  }

  ansi = X11DRV_MapVirtualKeyEx(vkey, 2, X11DRV_GetKeyboardLayout(0));
  TRACE_(key)("scan 0x%04x, vkey 0x%04x, ANSI 0x%04x\n", scanCode, vkey, ansi);

  /* first get the name of the "regular" keys which is the Upper case
     value of the keycap imprint.                                     */
  if ( ((ansi >= 0x21) && (ansi <= 0x7e)) &&
       (scanCode != 0x137) &&   /* PrtScn   */
       (scanCode != 0x135) &&   /* numpad / */
       (scanCode != 0x37 ) &&   /* numpad * */
       (scanCode != 0x4a ) &&   /* numpad - */
       (scanCode != 0x4e ) )    /* numpad + */
      {
        if ((nSize >= 2) && lpBuffer)
        {
          *lpBuffer = toupperW((WCHAR)ansi);
          *(lpBuffer+1) = 0;
          return 1;
        }
     else
        return 0;
  }

  /* FIXME: horrible hack to fix function keys. Windows reports scancode
            without "extended-key" flag. However Wine generates scancode
            *with* "extended-key" flag. Seems to occur *only* for the
            function keys. Soooo.. We will leave the table alone and
            fudge the lookup here till the other part is found and fixed!!! */

  if ( ((scanCode >= 0x13b) && (scanCode <= 0x144)) ||
       (scanCode == 0x157) || (scanCode == 0x158))
    scanCode &= 0xff;   /* remove "extended-key" flag for Fx keys */

  /* let's do scancode -> keycode -> keysym -> String */

  for (keyi=kbd.min_keycode; keyi<=kbd.max_keycode; keyi++)
      if ((kbd.keyc[kbd.layout_no]->scan[keyi]) == scanCode)
         break;
  if (keyi <= kbd.max_keycode)
  {
      wine_tsx11_lock();
      keyc = (KeyCode) keyi;
      keys = XKeycodeToKeysym(thread_display(), keyc, 0);
      name = XKeysymToString(keys);
      wine_tsx11_unlock();
      TRACE_(key)("found scan=%04x keyc=%04x keysym=%04x string=%s\n",
            scanCode, keyc, (int)keys, name);
      if (lpBuffer && nSize && name)
      {
          MultiByteToWideChar(CP_UNIXCP, 0, name, -1, lpBuffer, nSize);
          lpBuffer[nSize - 1] = 0;
          return 1;
      }
  }

  /* Finally issue WARN for unknown keys   */

  WARN("(%08x,%p,%d): unsupported key, vkey=%04x, ansi=%04x\n",lParam,lpBuffer,nSize,vkey,ansi);
  if (lpBuffer && nSize)
    *lpBuffer = 0;
  return 0;
}

/***********************************************************************
 *		X11DRV_KEYBOARD_MapDeadKeysym
 */
static char KEYBOARD_MapDeadKeysym(KeySym keysym)
{
	switch (keysym)
	    {
	/* symbolic ASCII is the same as defined in rfc1345 */
#ifdef XK_dead_tilde
	    case XK_dead_tilde :
#endif
	    case 0x1000FE7E : /* Xfree's XK_Dtilde */
		return '~';	/* '? */
#ifdef XK_dead_acute
	    case XK_dead_acute :
#endif
	    case 0x1000FE27 : /* Xfree's XK_Dacute_accent */
		return 0xb4;	/* '' */
#ifdef XK_dead_circumflex
	    case XK_dead_circumflex:
#endif
	    case 0x1000FE5E : /* Xfree's XK_Dcircumflex_accent */
		return '^';	/* '> */
#ifdef XK_dead_grave
	    case XK_dead_grave :
#endif
	    case 0x1000FE60 : /* Xfree's XK_Dgrave_accent */
		return '`';	/* '! */
#ifdef XK_dead_diaeresis
	    case XK_dead_diaeresis :
#endif
	    case 0x1000FE22 : /* Xfree's XK_Ddiaeresis */
		return 0xa8;	/* ': */
#ifdef XK_dead_cedilla
	    case XK_dead_cedilla :
	        return 0xb8;	/* ', */
#endif
#ifdef XK_dead_macron
	    case XK_dead_macron :
	        return '-';	/* 'm isn't defined on iso-8859-x */
#endif
#ifdef XK_dead_breve
	    case XK_dead_breve :
	        return 0xa2;	/* '( */
#endif
#ifdef XK_dead_abovedot
	    case XK_dead_abovedot :
	        return 0xff;	/* '. */
#endif
#ifdef XK_dead_abovering
	    case XK_dead_abovering :
	        return '0';	/* '0 isn't defined on iso-8859-x */
#endif
#ifdef XK_dead_doubleacute
	    case XK_dead_doubleacute :
	        return 0xbd;	/* '" */
#endif
#ifdef XK_dead_caron
	    case XK_dead_caron :
	        return 0xb7;	/* '< */
#endif
#ifdef XK_dead_ogonek
	    case XK_dead_ogonek :
	        return 0xb2;	/* '; */
#endif
/* FIXME: I don't know this three.
	    case XK_dead_iota :
	        return 'i';
	    case XK_dead_voiced_sound :
	        return 'v';
	    case XK_dead_semivoiced_sound :
	        return 's';
*/
	    }
	return 0;
}

/***********************************************************************
 *		ToUnicodeEx (X11DRV.@)
 *
 * The ToUnicode function translates the specified virtual-key code and keyboard
 * state to the corresponding Windows character or characters.
 *
 * If the specified key is a dead key, the return value is negative. Otherwise,
 * it is one of the following values:
 * Value Meaning
 * 0    The specified virtual key has no translation for the current state of the keyboard.
 * 1    One Windows character was copied to the buffer.
 * 2    Two characters were copied to the buffer. This usually happens when a
 *      dead-key character (accent or diacritic) stored in the keyboard layout cannot
 *      be composed with the specified virtual key to form a single character.
 *
 * FIXME : should do the above (return 2 for non matching deadchar+char combinations)
 *
 */
INT X11DRV_ToUnicodeEx(UINT virtKey, UINT scanCode, LPBYTE lpKeyState,
                     LPWSTR bufW, int bufW_size, UINT flags, HKL hkl)
{
    Display *display = thread_display();
    XKeyEvent e;
    KeySym keysym = 0;
    INT ret;
    int keyc, savetype;
    unsigned int savestate;
    char lpChar[10];
    HWND focus;
    XIC xic;
    Status status;
    int layout_no = kbd.layout_no;

    if (scanCode & 0x8000)
    {
        TRACE_(key)("Key UP, doing nothing\n" );
        return 0;
    }

    if (hkl != (HKL)(main_key_tab[kbd.layout[layout_no]].lcid)) {
#ifdef HAVE_XKB 
        for(layout_no=0; layout_no<kbd.num_layouts; layout_no++) {
           if (!kbd.pref[layout_no]) continue;
           if (hkl == (HKL)(main_key_tab[kbd.layout[layout_no]].lcid)) break;
        }
    }
    if (layout_no == kbd.num_layouts) {
        layout_no = kbd.layout_no;
#endif /* HAVE_XKB */
        TRACE_(key)("keyboard layout %p is not supported\n", hkl);
    }

    if ((lpKeyState[VK_MENU] & 0x80) && (lpKeyState[VK_CONTROL] & 0x80))
    {
        TRACE_(key)("Ctrl+Alt+[key] won't generate a character\n");
        return 0;
    }

    e.display = display;
    e.keycode = 0;
    e.state = 0;
    e.type = KeyPress;

    focus = GetFocus();
    if (focus) focus = GetAncestor( focus, GA_ROOT );
    if (!focus) focus = GetActiveWindow();
    e.window = X11DRV_get_whole_window( focus );
    xic = X11DRV_get_ic( focus );

    if (lpKeyState[VK_SHIFT] & 0x80)
    {
        TRACE_(key)("ShiftMask = %04x\n", ShiftMask);
        e.state |= ShiftMask;
    }
    if (lpKeyState[VK_CAPITAL] & 0x01)
    {
        TRACE_(key)("LockMask = %04x\n", LockMask);
        e.state |= LockMask;
    }
    if (lpKeyState[VK_CONTROL] & 0x80)
    {
        TRACE_(key)("ControlMask = %04x\n", ControlMask);
        e.state |= ControlMask;
    }
    if (lpKeyState[VK_NUMLOCK] & 0x01)
    {
        TRACE_(key)("NumLockMask = %04x\n", NumLockMask);
        e.state |= NumLockMask;
    }

    /* Restore saved AltGr state */
    TRACE_(key)("AltGrMask = %04x\n", AltGrMask);
    e.state |= AltGrMask;

    TRACE_(key)("(%04X, %04X) : faked state = 0x%04x\n",
                virtKey, scanCode, e.state);
    wine_tsx11_lock();
    /* We exit on the first keycode found, to speed up the thing. */
    for (keyc=kbd.min_keycode; (keyc<=kbd.max_keycode) && (!e.keycode) ; keyc++)
      { /* Find a keycode that could have generated this virtual key */
          if  ((kbd.keyc[layout_no]->vkey[keyc] & 0xFF) == virtKey)
          { /* We filter the extended bit, we don't know it */
              e.keycode = KeycToOver(keyc); /* Store it temporarily */
              if ( keyc != e.keycode )
                  TRACE_(key)("Keycode %02x turned back to %02x\n", keyc, e.keycode);
              if ((EVENT_event_to_vkey(xic,&e) & 0xFF) != virtKey) {
                  e.keycode = 0; /* Wrong one (ex: because of the NumLock
                         state), so set it to 0, we'll find another one */
              }
              else {
                e.keycode = KeycToOver(keyc); /* again */
              }
          }
      }

    if (virtKey >= VK_LEFT && virtKey <= VK_DOWN)
        e.keycode = XKeysymToKeycode(e.display, virtKey - VK_LEFT + XK_Left);

    if ((virtKey>=VK_NUMPAD0) && (virtKey<=VK_NUMPAD9))
        e.keycode = XKeysymToKeycode(e.display, virtKey-VK_NUMPAD0+XK_KP_0);

    if (virtKey==VK_DECIMAL)
        e.keycode = XKeysymToKeycode(e.display, XK_KP_Decimal);

    if (virtKey==VK_SEPARATOR)
        e.keycode = XKeysymToKeycode(e.display, XK_KP_Separator);

    if (!e.keycode && virtKey != VK_NONAME)
      {
        WARN_(key)("Unknown virtual key %X !!!\n", virtKey);
        wine_tsx11_unlock();
        return 0;
      }
    else TRACE_(key)("Found keycode %d (0x%2X)\n",e.keycode,e.keycode);

    TRACE_(key)("type %d, window %lx, state 0x%04x, keycode 0x%04x\n",
                e.type, e.window, e.state, e.keycode);

    savestate = e.state;
    if (xic) {
          savetype = e.type;
          if (savetype == KeyRelease) e.type = KeyPress;
          ret = XmbLookupString(xic, &e, lpChar, sizeof(lpChar), &keysym, &status);
          e.type = savetype;
        }
    else
        ret = XLookupString(&e, lpChar, sizeof(lpChar), &keysym, NULL);
    e.state = savestate;
    wine_tsx11_unlock();

#ifdef XK_EuroSign
    /* An ugly hack for EuroSign: X can't translate it to a character
       for some locales. */
    if (keysym == XK_EuroSign)
    {
        bufW[0] = 0x20AC;
        ret = 1;
        goto found;
    }
#endif

    if ((keysym > 0) && (keysym < 0x8000) && kbd.cp[layout_no]) {
    /* Get lower byte */
        lpChar[0] = keysym & 0xFF;
        if (MultiByteToWideChar(kbd.cp[layout_no], 0, lpChar, 1, bufW, bufW_size)) {
            ret = 1;
        } else {
            TRACE_(key)("keysym 0x%04lx treated as char 0x%02x from codepage %d, but conversion to Unicode failed\n",
                keysym,lpChar[0],kbd.cp[layout_no]);
            ret = 0;
        }
    }

    if (ret == 0)
    {
        char dead_char;

        /* Special case: X turns shift-tab into ISO_Left_Tab. */
        /* Here we change it back. */
        if (keysym == XK_ISO_Left_Tab)
        {
            bufW[0] = 0x09;
            ret = 1;
            goto found;
        }

        dead_char = KEYBOARD_MapDeadKeysym(keysym);
        if (dead_char)
        {
            if (kbd.cp[layout_no])
                MultiByteToWideChar(kbd.cp[layout_no], 0, &dead_char, 1, bufW, bufW_size);
            else
                MultiByteToWideChar(CP_UNIXCP, 0, &dead_char, 1, bufW, bufW_size);
            ret = -1;
            goto found;
        } else {
            if (keysym >= 0x01000100 && keysym <= 0x0100ffff)
            {
                /* Unicode direct mapping */
                bufW[0] = keysym & 0xffff;
                ret = 1;
                goto found;
            }
            else if ((keysym >> 8) == 0x1008FF) {
                bufW[0] = 0;
                ret = 0;
                goto found;
            }
            TRACE_(key)("MapDeadKeysym failed for dead keysym 0x%08lx\n",keysym);
            const char *ksname;

            wine_tsx11_lock();
            ksname = XKeysymToString(keysym);
            wine_tsx11_unlock();
            if (!ksname)
                ksname = "No Name";
            if ((keysym >> 8) != 0xff)
                {
                WARN("Please report: no char for keysym %04lX (%s) :\n",
                    keysym, ksname);
                WARN("(virtKey=%X,scanCode=%X,keycode=%X,state=%X)\n",
                    virtKey, scanCode, e.keycode, e.state);
                }
            goto found;
        }
    } else {    
    /* ret != 0 */
        /* We have a special case to handle : Shift + arrow, shift + home, ...
           X returns a char for it, but Windows doesn't. Let's eat it. */
        if (!(e.state & NumLockMask)  /* NumLock is off */
            && (e.state & ShiftMask) /* Shift is pressed */
            && (keysym>=XK_KP_0) && (keysym<=XK_KP_9))
        {
            lpChar[0] = 0;
            ret = 0;
        }

        /* more areas where X returns characters but Windows does not
           CTRL + number or CTRL + symbol */
        if (e.state & ControlMask)
        {
            if (((keysym>=33) && (keysym < 'A')) ||
                ((keysym > 'Z') && (keysym < 'a')))
            {
                lpChar[0] = 0;
                ret = 0;
            }
        }

        /* We have another special case for delete key (XK_Delete) on an
         extended keyboard. X returns a char for it, but Windows doesn't */
        if (keysym == XK_Delete)
        {
            lpChar[0] = 0;
            ret = 0;
        }
        else if((lpKeyState[VK_SHIFT] & 0x80) /* Shift is pressed */
                && (keysym == XK_KP_Decimal))
        {
            lpChar[0] = 0;
            ret = 0;
        }

        if(ret) {
        /* perform translation to unicode */
            {
                if (kbd.cp[layout_no]) {
                    TRACE_(key)("Translating char 0x%02x from codepage %d to unicode\n", 
                        *(BYTE *)lpChar,kbd.cp[layout_no]);
                    ret = MultiByteToWideChar(kbd.cp[layout_no], 0, lpChar, 1, bufW, bufW_size);
                } else 
                {
                    TRACE_(key)("Translating char 0x%02x to unicode\n", *(BYTE *)lpChar);
                    ret = MultiByteToWideChar(CP_UNIXCP, 0, lpChar, ret, bufW, bufW_size);
                }
            }
        }
    }

found:
    TRACE_(key)("ToUnicode about to return %d with char %x %s\n",
                ret, (ret && bufW) ? bufW[0] : 0, bufW ? "" : "(no buffer)");
    return ret;
}

/***********************************************************************
 *		Beep (X11DRV.@)
 */
void X11DRV_Beep(void)
{
    wine_tsx11_lock();
    XBell(thread_display(), 0);
    wine_tsx11_unlock();
}
